# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from gi.repository import Gio, GLib, GObject, Gtk

from hashlib import sha256
import logging
from threading import Lock, Thread
import time
from typing import Optional

from iotas.attachment import Attachment
from iotas.attachment_helpers import write_attachment_to_disk
from iotas.category_manager import CategoryManager
from iotas.config_manager import ConfigManager
from iotas.nextcloud_sync_worker import NextcloudSyncWorker, LoginRequestResult
from iotas.note import Note
from iotas.note_database import NoteDatabase
from iotas.note_manager import NoteManager
from iotas.sync_result import ContentPushSyncResult, FailedPush
from iotas.ui_utils import idle_add_wait


class SyncWorkLock:

    def __init__(self) -> None:
        self.__lock = Lock()
        self.__locked = False

    def acquire(self) -> None:
        """Acquire the lock."""
        self.__locked = True
        self.__lock.acquire()

    def release(self) -> None:
        """Release the lock."""
        self.__locked = False
        self.__lock.release()

    @property
    def locked(self) -> bool:
        return self.__locked


class SyncManager(GObject.Object):
    __gsignals__ = {
        # bool: Whether a new account
        "ready": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "started": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "have-changes": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # int: Number of changes
        "finished": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # Note: Note in conflict
        "note-conflict": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        # Note: Remotely deleted note
        "note-deleted": (GObject.SignalFlags.RUN_FIRST, None, (Note,)),
        "secret-service-failure": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # str: Previous category, str: Remote category
        "remote-category-sanitised": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        "missing-password": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "notes-capability-missing": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "attachment-downloaded": (GObject.SignalFlags.RUN_FIRST, None, (Attachment,)),
    }

    STARTUP_SYNC_INIT_DURATION = 500
    SYNC_FAIL_OFFLINE_THRESHOLD = 2
    MIN_NEW_NOTES_FOR_BATCHING = 30
    SYNC_BATCH_SIZE = 5

    def __init__(self, db: NoteDatabase) -> None:
        super().__init__()
        self.__db = db
        self.__config_manager = ConfigManager.get_default()
        self.__sync_worker = NextcloudSyncWorker(self.__config_manager)
        self.__lock = SyncWorkLock()

        self.__note_manager: NoteManager
        self.__category_manager: CategoryManager

        self.__authenticated = False
        self.__api_check_complete = False
        self.__initial_sync_delay_complete = False
        self.__sync_requested_while_active = False
        self.__waiting_timeout = None
        self.__offline = False
        self.__sync_fail_counter = 0

        self.__config_manager.connect_changed(
            ConfigManager.SYNC_INTERVAL,
            self.__set_sync_interval,
        )
        self.__set_sync_interval()

        self.__config_manager.bind(
            ConfigManager.NETWORK_TIMEOUT,
            self.__sync_worker,
            "timeout",
            Gio.SettingsBindFlags.DEFAULT,
        )

        self.__sync_worker.connect("ready", self.__on_sync_ready)
        self.__sync_worker.connect("api-check-finished", self.__on_api_check_finished)
        self.__sync_worker.connect("settings-fetched", self.__on_settings_fetched)
        self.__sync_worker.connect(
            "secret-service-failure", lambda _o: GLib.idle_add(self.emit, "secret-service-failure")
        )
        self.__sync_worker.bind_property(
            "authenticated", self, "authenticated", GObject.BindingFlags.SYNC_CREATE
        )
        self.__sync_worker.connect("missing-password", self.__on_missing_password)
        self.__sync_worker.connect(
            "notes-capability-missing", lambda _o: self.emit("notes-capability-missing")
        )

    def init_auth(self) -> None:
        """Initialise sync authentication details from the Secret Service."""
        self.__sync_worker.init_auth()

    def set_managers(self, note_manager: NoteManager, category_manager: CategoryManager) -> None:
        """Set the models.

        :param NoteManager note_manager: The note manager
        :param CategoryManager category_manager: The category manager
        """
        self.__note_manager = note_manager
        self.__note_manager.connect(
            "initial-load-complete", lambda _o: self.__on_initial_note_load_complete()
        )
        self.__category_manager = category_manager

    def sync_now(self) -> None:
        """Attempt to initiate a sync.

        If the server isn't authenticated the request will be discarded. Another sync already
        underway will result in the sync being run immediately after the current operation.
        """
        if not self.__sync_worker.authenticated:
            return

        if self.__lock.locked:
            self.__sync_requested_while_active = True
            return

        if self.__waiting_timeout:
            GLib.source_remove(self.__waiting_timeout)
            self.__waiting_timeout = None

        self.__sync()

    def close(self) -> None:
        """Close the manager, perform cleanup."""
        self.__config_manager.nextcloud_prune_threshold = self.__sync_worker.prune_threshold

    def reset_marker(self) -> None:
        """Reset the most recent sync timestamp market."""
        self.__sync_worker.reset_prune_threshold()

    def sign_out(self) -> None:
        """Sign out from the sync server."""
        self.__sync_worker.sign_out()

    def flush_pending_deletions(self) -> None:
        """Flush any notes awaiting their pending undo period."""
        notes_pending_undo = self.__note_manager.get_deleted_notes_pending_undo()
        if notes_pending_undo and self.authenticated:
            for note in notes_pending_undo:
                if note.has_remote_id:
                    self.sync_now()
                    break
        self.__note_manager.set_deleted_notes_pending_undo([])

    def start_login(
        self, server_uri: str, window: Gtk.Window
    ) -> tuple[LoginRequestResult, Optional[str], Optional[str]]:
        """Initiate a new login.

        :param str server_uri: Server instance URI
        :param Gtk.Window window: Main window
        :return: Tuple with LoginRequestResult before endpoint URI and auth token (or Nones on
            failures)
        :rtype: tuple[LoginRequestResult, Optional[str], Optional[str]]
        """
        return self.__sync_worker.start_login(server_uri, window)

    def check_for_login_success(self, endpoint: str, token: str) -> tuple[bool, bool]:
        """Check whether login succeeded.

        :param str endpoint: Endpoint URI
        :param str token: Request token
        :return: Whether authenticated and whether auth. storage succeeded
        :rtype: tuple[bool, bool]
        """
        return self.__sync_worker.check_endpoint_for_auth_token(endpoint, token)

    def download_attachment(self, attachment: Attachment) -> bool:
        """Fetch attachment from server and store.

        :param Attachment attachment: Attachment to transfer
        :return: Success
        :rtype: bool
        """
        self.__check_and_warn_for_other_note_attachment(attachment)
        result = self.__sync_worker.get_attachment(attachment)
        if not result.data:
            note = self.__note_manager.fetch_note_by_id(attachment.note_id)
            title = note.title if note else "?"  # mypy
            if result.status_code == 404:
                logging.warning(
                    f"Failed to download missing attachment '{attachment.path}' for note "
                    f"'{title}' ({result.status_code})"
                )
            else:
                logging.warning(
                    f"Failed to download attachment '{attachment.path}' for note "
                    f"'{title}' ({result.status_code})"
                )

            return False

        if not self.__db.get_note_by_remote_id(attachment.note_remote_id):
            logging.warning("Discarding attachment for removed note")
            # Returning true as this isn't a failure
            return True

        if not write_attachment_to_disk(attachment, result.data):
            return False

        GLib.idle_add(self.emit, "attachment-downloaded", attachment)
        return True

    @GObject.Property(type=bool, default=False)
    def authenticated(self) -> bool:
        return self.__authenticated

    @authenticated.setter
    def set_authenticated(self, new_val: bool) -> None:
        self.__authenticated = new_val

    @GObject.Property(type=bool, default=False)
    def configured_but_no_password(self) -> bool:
        return (
            self.__config_manager.nextcloud_endpoint != ""
            and self.__config_manager.nextcloud_username != ""
            and not self.authenticated
        )

    @GObject.Property(type=bool, default=False)
    def offline(self) -> bool:
        return self.__offline

    @offline.setter
    def set_offline(self, new_val: bool) -> None:
        self.__offline = new_val

    def __on_initial_note_load_complete(self) -> None:
        if self.__config_manager.nextcloud_username != "":
            GLib.timeout_add(self.STARTUP_SYNC_INIT_DURATION, self.__perform_initial_sync)

    def __sync(self) -> None:
        self.__waiting_timeout = None

        if self.__lock.locked:
            self.__sync_requested_while_active = True
            return

        GLib.idle_add(self.emit, "started")
        thread = Thread(target=self.__sync_on_thread)
        thread.daemon = True
        thread.start()

    def __sync_on_thread(self) -> None:
        logging.debug("Refresh....")

        self.__lock.acquire()
        self.__sync_requested_while_active = False

        start_time = time.time()
        fetching_all = self.__sync_worker.prune_threshold == 0
        sync_result = self.__sync_worker.get_notes()
        if sync_result.success:
            logging.debug("Server GET took {:.2f}s".format(time.time() - start_time))
        else:
            logging.info(f"Refresh failed with status code {sync_result.status_code}")

            # Ensure our API check gets run, just in case a major API version issue is causing
            # the issue
            self.__api_check_and_settings_sync()

            self.__finish_sync(0)
            self.__sync_fail_counter += 1
            if not self.__offline and self.__sync_fail_counter >= self.SYNC_FAIL_OFFLINE_THRESHOLD:
                logging.info("Flagging offline")
                self.offline = True
            return
        self.__sync_fail_counter = 0
        if self.__offline:
            logging.info("Recovered connectivity")
            self.offline = False

        notes = sync_result.data
        assert notes is not None

        remote_ids_from_db = self.__db.get_all_notes_remote_ids()
        # logging.debug("Pulled {} notes".format(len(notes)))

        # Determine remote creations and updates
        remote_new, remote_upd = self.__determine_remote_new_and_updated(
            notes, remote_ids_from_db, fetching_all
        )

        # Remove duplicate updates caused by eg. ignored No-Cache headers
        remote_upd = self.__filter_out_false_updates(remote_upd)

        # Pre-processing for remote updates to notes which have local changes
        remote_upd = self.__preprocess_remote_updates_for_locally_modified(remote_upd)

        # Determine any remote deletions
        sync_server_remote_ids = [v["id"] for v in notes]
        remote_del = list(set(remote_ids_from_db) - set(sync_server_remote_ids))

        # Check for delete conflicts
        self.__check_for_remote_delete_conflicts(remote_del)

        # Determine, sync and locally apply local deletions
        local_del = self.__apply_local_deletions()

        # Determine and sync local updates
        local_upd = self.__apply_local_updates()

        # Determine and sync local creations
        local_new = self.__apply_local_creations()

        changes_out = len(local_new + local_upd + local_del)
        if changes_out > 0:
            logging.debug(
                "Changes OUT: %s new %s upd %s del",
                len(local_new),
                len(local_upd),
                len(local_del),
            )
        changes_in = len(remote_new + remote_upd + remote_del)
        if changes_in > 0:
            logging.debug(
                "Changes IN: %s new %s upd %s del",
                len(remote_new),
                len(remote_upd),
                len(remote_del),
            )
        if changes_in > 0:
            idle_add_wait(self.emit, "have-changes")

            # Integrate remote changes, batching for large initial transfers
            if len(remote_new) > self.MIN_NEW_NOTES_FOR_BATCHING:
                batch_size = self.SYNC_BATCH_SIZE
            else:
                batch_size = self.MIN_NEW_NOTES_FOR_BATCHING
            for i in range(0, max(1, len(remote_new)), batch_size):
                if len(remote_new) > 0:
                    batch = remote_new[i : i + batch_size]
                else:
                    batch = []
                idle_add_wait(self.__integrate_remote_changes, batch, remote_upd, remote_del)
                remote_upd = []
                remote_del = []

        # API check and settings sync are slung onto the back of the first request so that first
        # sync isn't as delayed. Justified by the fact that the outcome of the API check isn't
        # currently impacting behaviour; this may move.
        self.__api_check_and_settings_sync()

        self.__finish_sync(changes_in + changes_out)

    def __determine_remote_new_and_updated(
        self, notes: list[dict], remote_ids_from_db: list[int], force_refresh_all: bool
    ) -> tuple[list[dict], list[dict]]:
        create = []
        update = []
        for sync_note in notes:
            # Notes before the prune cutoff will only have an "id" item
            if len(sync_note) == 1:
                continue

            if sync_note["id"] not in remote_ids_from_db:
                create.append(sync_note)
            elif force_refresh_all or self.__db.note_needs_update(
                sync_note["id"], sync_note["etag"]
            ):
                update.append(sync_note)

        return (create, update)

    def __filter_out_false_updates(self, updates_in: list[dict]) -> list[dict]:
        updates_out = []
        for remote_note_dict in updates_in:
            # Fetch note
            local_note = self.__note_manager.fetch_note_by_remote_id(remote_note_dict["id"])
            if local_note is None:
                logging.warning(
                    f"Failed to find local note with remote id '{remote_note_dict['id']}'"
                )
                continue

            # Handle possible false updates from misbehaving proxies etc
            remote_note = Note.from_nextcloud(remote_note_dict)
            if local_note.etag_is_recent(remote_note.etag):
                if local_note.identical_non_sync_meta(remote_note):
                    # The update from the server is an ETag from a recent push with no metadata
                    # changes, we're likely receiving cached or otherwise old content. Ignore the
                    # update.
                    logging.debug(
                        f"Remote note appears to be a past push, etag '{remote_note.etag}'"
                    )
                    # Skip adding for further processing
                    continue
                else:
                    if not local_note.content_loaded:
                        self.__db.populate_note_content(local_note)
                    if local_note.dirty and local_note.content == remote_note.content:
                        # If our local note has changed for something other than content we
                        # possibly have a real conflict. Allow falling through to that.
                        logging.debug(
                            "Note update received for locally modified note with no etag change "
                            "and no content change"
                        )
                    else:
                        # Possibly just a remote metadata change, but let's ignore the
                        # content to be safe
                        logging.debug("Note received with past etag but meta update")
                        remote_note_dict["ignore-content"] = True

            updates_out.append(remote_note_dict)

        return updates_out

    def __preprocess_remote_updates_for_locally_modified(
        self, updates_in: list[dict]
    ) -> list[dict]:
        updates_out = []
        for remote_note_dict in updates_in:
            # Fetch note
            local_note = self.__note_manager.fetch_note_by_remote_id(remote_note_dict["id"])
            if local_note is None:
                logging.warning(
                    f"Failed to find local note with remote id '{remote_note_dict['id']}'"
                )
                continue

            # Duplicate filtering can flag notes for which content changes should be ignored
            ignore_content_change = "ignore-content" in remote_note_dict

            # Handle local note with changes
            if local_note.dirty and not ignore_content_change:
                remote_note = Note.from_nextcloud(remote_note_dict)
                new_dirty = False
                if not local_note.content_loaded:
                    self.__db.populate_note_content(local_note)
                if local_note.identical_excepting_sync_meta(remote_note):
                    logging.info("Applying remote sync meta to otherwise identical note")
                elif self.__conflict_is_recent_failed_push(local_note, remote_note):
                    # The remote note is an older version of the local note. We attempted to sync
                    # it but didn't receive an ack and as a result our local etag is out of date.
                    # So we update the etag and ignore everything else.
                    logging.info(
                        "Avoiding conflict as remote note with updated etag appears to be "
                        "a failed past push"
                    )
                    local_note.etag = remote_note.etag
                    self.__db.persist_note_etag(local_note)
                    # Skip adding for further processing
                    continue
                else:
                    logging.info(
                        f"Sync. conflict, note '{local_note.title}' was modified locally and "
                        "updated remotely"
                    )
                    GLib.idle_add(self.emit, "note-conflict", local_note)
                    # Duplicate existing note
                    GLib.idle_add(self.__duplicate_note_for_sync_conflict, local_note)

                local_note.dirty = new_dirty
                self.__db.persist_note_dirty(local_note)

            updates_out.append(remote_note_dict)

        return updates_out

    def __check_for_remote_delete_conflicts(self, remote_del: list[int]) -> None:
        for remote_id in remote_del:
            note = self.__note_manager.fetch_note_by_remote_id(remote_id)
            if note is None:
                continue
            if note.dirty and not note.locally_deleted:
                GLib.idle_add(self.emit, "note-conflict", note)
                # Duplicate existing note
                logging.info(
                    "Sync. conflict, note '%s' was modified locally and deleted remotely",
                    note.title,
                )
                GLib.idle_add(self.__duplicate_note_for_sync_conflict, note)
                note.dirty = False
                self.__db.persist_note_dirty(note)

    def __apply_local_updates(self) -> list[Note]:
        local_upd = [
            v
            for v in self.__note_manager.base_model
            if v.dirty and not v.locally_deleted and v.remote_id >= 0
        ]
        for note in local_upd:
            if not note.content_loaded:
                self.__db.populate_note_content(note)
            note.saving = True
            res = self.__sync_worker.update_note(note)
            idle_add_wait(self.__update_note_metadata_post_sync, note, res)
        return local_upd

    def __apply_local_creations(self) -> list[Note]:
        local_new = [
            n
            for n in self.__note_manager.base_model
            if n.has_id and n.dirty and not n.locally_deleted and n.remote_id < 0
        ]

        for note in local_new:
            if not note.content_loaded:
                self.__db.populate_note_content(note)
            note.saving = True
            res = self.__sync_worker.create_note(note)
            idle_add_wait(self.__update_note_metadata_post_sync, note, res)
        return local_new

    def __apply_local_deletions(self) -> list[Note]:
        local_del_queue = [
            n
            for n in self.__note_manager.base_model
            if n.locally_deleted and n not in self.__note_manager.get_deleted_notes_pending_undo()
        ]
        local_del = []
        for note in local_del_queue:
            # logging.debug("Processing {} in local_del_queue".format(len(local_del_queue)))
            if note.remote_id >= 0:
                res = self.__sync_worker.delete_note(note)
                if res.success:
                    local_del.append(note)
            else:
                local_del.append(note)
        if len(local_del) > 0:
            for note in local_del:
                self.__db.delete_note(note.id)
            idle_add_wait(self.__note_manager.remove_notes_from_model, local_del)
        return local_del

    def __finish_sync(self, number_of_changes: int) -> None:
        GLib.idle_add(self.emit, "finished", number_of_changes)
        if self.__sync_requested_while_active:
            self.__lock.release()
            self.__sync()
        else:
            self.__waiting_timeout = GLib.timeout_add_seconds(self.__sync_interval, self.__sync)
            self.__lock.release()

    def __integrate_remote_changes(
        self, remote_new: list[dict], remote_upd: list[dict], remote_del: list[int]
    ) -> None:
        for sync_note in remote_new + remote_upd:
            existing_note = self.__note_manager.fetch_note_by_remote_id(sync_note["id"])
            if existing_note:
                if not existing_note.content_loaded:
                    self.__db.populate_note_content(existing_note)
                old_category = existing_note.category
                updated_fields = existing_note.update_from_remote_sync(sync_note)
                # Updated fields can be empty when doing a schema migration full resync
                if updated_fields.empty():
                    continue
                self.__db.persist_note_selective(existing_note, updated_fields)
                if updated_fields.category:
                    self.__category_manager.note_category_changed(
                        old_category, existing_note.category
                    )
            else:
                note = Note.from_nextcloud(sync_note)
                self.__db.add_note(note)
                self.__note_manager.add_notes_to_model([note])
                self.__category_manager.note_added(note.category)

        remove_notes = []
        for remote_id in remote_del:
            deleted_note = self.__note_manager.fetch_note_by_remote_id(remote_id)
            if deleted_note is not None:
                self.emit("note-deleted", deleted_note)
                remove_notes.append(deleted_note)
        if len(remove_notes) > 0:
            self.__note_manager.delete_notes(remove_notes, provide_undo=False)

    def __duplicate_note_for_sync_conflict(self, note: Note) -> None:
        if not note.content_loaded:
            self.__db.populate_note_content(note)
        # Translators: Description, used as a prefix to the previous title for notes updated both
        # locally and remotely. " - " is placed between this prefix and the title.
        reason = _("SYNC CONFLICT")
        new_note = self.__db.create_duplicate_note(note, reason)
        self.__note_manager.add_notes_to_model([new_note])
        self.__category_manager.note_added(new_note.category)

        new_note.saving = True
        res = self.__sync_worker.create_note(new_note)
        self.__update_note_metadata_post_sync(new_note, res)

    def __update_note_metadata_post_sync(
        self, note: Note, sync_result: ContentPushSyncResult
    ) -> None:
        if sync_result.success:
            previous_category = note.category
            remote_category = sync_result.data["category"]
            self.__note_manager.update_note_post_sync(note, sync_result)
            if previous_category != remote_category:
                GLib.idle_add(
                    self.emit, "remote-category-sanitised", previous_category, remote_category
                )
        else:
            logging.info(f"Note push failed (with status code {sync_result.status_code})")

            if sync_result.sent_content is not None:
                # Track this failed push so that it can be used to avoid possible conflicts if the
                # push actually did make it to the server but we haven't received the tag to update.
                self.__track_failed_push(note, sync_result.sent_content)
            else:
                logging.warning("Wanted to track failed push with null sent content")

        note.saving = False
        if not sync_result.success or note.dirty_while_saving:
            note.flag_changed(update_last_modified=False)
            note.dirty_while_saving = False

    def __on_sync_ready(self, _obj: GObject.Object, new_setup: bool) -> None:
        GLib.idle_add(self.emit, "ready", new_setup)
        if new_setup or self.__initial_sync_delay_complete:
            self.sync_now()

    def __on_missing_password(self, _obj: GObject.Object) -> None:
        if self.configured_but_no_password:
            logging.info(
                "Sync is configured but password missing. Login to the server to reauthenticate."
            )
        self.emit("missing-password")

    def __on_api_check_finished(self, _obj: GObject.Object) -> None:
        self.__api_check_complete = True
        thread = Thread(target=self.__sync_worker.fetch_settings)
        thread.daemon = True
        thread.start()

    def __on_settings_fetched(self, _obj: GObject.Object, file_suffix: str) -> None:
        if self.__config_manager.backup_note_extension != file_suffix:
            logging.info(
                f"Changed backup note filename extension to '{file_suffix}' "
                "as received from server"
            )
            self.__config_manager.backup_note_extension = file_suffix

    def __set_sync_interval(self) -> None:
        self.__sync_interval = self.__config_manager.sync_interval

    def __api_check_and_settings_sync(self) -> None:
        # API check and settings sync are slung onto the back of the first request so that first
        # sync isn't as delayed. Justified by the fact that the outcome of the API check isn't
        # currently impacting behaviour.
        if not self.__api_check_complete:
            thread = Thread(target=self.__sync_worker.check_capabilities)
            thread.daemon = True
            thread.start()

    def __perform_initial_sync(self) -> None:
        self.__initial_sync_delay_complete = True
        self.sync_now()

    def __track_failed_push(self, note: Note, sent_content: str) -> None:
        hash = sha256(sent_content.encode("utf-8")).hexdigest()
        length = len(sent_content)
        timestamp = int(time.time())
        push = FailedPush(hash, length, timestamp)
        self.__note_manager.add_failed_push(note, push)

    def __conflict_is_recent_failed_push(self, note: Note, remote_note: Note) -> bool:
        pushes = self.__note_manager.get_note_failed_pushes(note)
        if not pushes:
            return False

        hash = sha256(remote_note.content.encode("utf-8")).hexdigest()
        length = len(remote_note.content)
        matched = False

        for push in pushes:
            if push.hash == hash and push.length == length:
                logging.info("Changed etag matches failed push, conflict avoided")
                matched = True
                break

        return matched

    def __check_and_warn_for_other_note_attachment(self, attachment: Attachment) -> None:
        path = attachment.path
        this_note_prefix = f".attachments.{attachment.note_remote_id}/"
        if path.startswith(".attachments.") and not path.startswith(this_note_prefix):
            logging.warning(
                f"Attachment '{attachment.filename}' appears to belong to another note. "
                "This is unsupported; unpredictable behaviour ahead."
            )



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject, Gtk

from iotas.category import Category


class CategoryTreeViewListStore(Gtk.ListStore):

    def __init__(self) -> None:
        super().__init__(GObject.TYPE_STRING, Category)

    def populate(self, categories: list[Category]) -> None:
        """Populate store with categories.

        :param list[Category] categories: Categories to populate
        """
        for category in categories:
            iter = self.append()
            self.set_value(iter, 0, category.display_name)
            self.set_value(iter, 1, category)

    def add(self, category: Category) -> None:
        """Add category to store.

        :param Category category: Category to remove
        """
        iter = self.get_iter_first()
        if iter is None:
            new_iter = self.append()
        else:
            while iter is not None:
                if self.get_value(iter, 1).name.lower() > category.name.lower():
                    break
                iter = self.iter_next(iter)
            new_iter = self.insert_before(iter)
        self.set_value(new_iter, 0, category.display_name)
        self.set_value(new_iter, 1, category)

    def delete(self, category: Category) -> None:
        """Remove category from store.

        :param Category category: Category to remove
        """
        iter = self.get_iter_first()
        while iter is not None:
            if self.get_value(iter, 1).name == category.name:
                self.remove(iter)
                break
            iter = self.iter_next(iter)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import logging
from requests import Response
from requests.exceptions import JSONDecodeError
from typing import Any, NamedTuple, Optional


class ResponseParseException(Exception):
    """Generic exception parsing response."""


class SyncResult(NamedTuple):

    success: bool
    status_code: Optional[int] = None
    data: Optional[list[dict]] = None

    @staticmethod
    def from_requests_response(response: Optional[Response]) -> SyncResult:
        """Create SyncResult from requests response.

        :param Optional[Response] response: The response
        :return: Object containing sync results
        :rtype: Self
        """
        try:
            success, response_json = parse_response(response)
        except ResponseParseException:
            success = False
            return SyncResult(success)
        if not isinstance(response_json, list):
            logging.warning(f"Data from server unexpected type {type(response_json)}")
            success = False
            return SyncResult(success)

        return SyncResult(success, response.status_code, response_json)


class ContentPushSyncResult(NamedTuple):
    success: bool
    status_code: Optional[int]
    data: Any
    sent_content: Optional[str]

    @staticmethod
    def from_requests_response(
        response: Optional[Response], sent_content: Optional[str]
    ) -> ContentPushSyncResult:
        """Create ContentPushSyncResult from requests response and sent content.

        The sent content is included here so it can be used in retaining a hash for failed pushes
        (which assist with false sync conflicts on high packet loss connections).

        :param Optional[Response] response: The response
        :param Optional[str] sent_content: The content as sent in the request
        :return: Object containing sync results and sent content
        :rtype: Self
        """
        try:
            success, response_json = parse_response(response)
        except ResponseParseException:
            status_code = None if not response else response.status_code
            return ContentPushSyncResult(False, status_code, None, sent_content)

        assert response is not None  # mypy

        return ContentPushSyncResult(success, response.status_code, response_json, sent_content)


class FailedPush(NamedTuple):
    hash: str
    length: int
    timestamp: int


def parse_response(response: Optional[Response]) -> tuple[bool, dict]:
    """Parse requests response.

    :param Optional[Response] response: The response
    :return: Whether the request was a success and the parsed response JSON
    :rtype: tuple[bool, dict]
    :raises ResponseParseException: If the response could not be parsed
    """
    if response is None:
        logging.debug("parse_response received null response")
        raise ResponseParseException("No response")

    try:
        response_json = response.json()
    except JSONDecodeError as e:
        logging.warning(f"Failed to decode server JSON response: {e}")
        raise ResponseParseException("Failed to decode JSON")

    success = response.status_code == 200

    return (success, response_json)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk, Gio

import logging
from typing import Any, Optional

from iotas.category import Category
from iotas.note import Note


class NoteListModelBase(Gtk.SortListModel):

    def __init__(self) -> None:
        self.__store_note_id_map: dict[int, Note] = {}
        self.__store_remote_id_map: dict[int, Note] = {}

        self.__list_store = Gio.ListStore(item_type=Note)
        self.__sorter = Gtk.CustomSorter()
        self.invalidate_sort()

        super().__init__(model=self.__list_store, sorter=self.__sorter)

    def invalidate_sort(self) -> None:
        """Invalidate the sorter."""
        self.__sorter.set_sort_func(NoteListModelBase.__sort_func)

    def add_notes(self, notes: list[Note]) -> None:
        """Add new notes to store.

        :param list[Note] notes: The notes to add
        """
        for note in notes:
            self.__list_store.append(note)
            if note.id != -1:
                self.__store_note_id_map[note.id] = note
            else:
                logging.warning("add_notes seeing -1 note id")
                continue
            if note.has_remote_id:
                self.__store_remote_id_map[note.remote_id] = note

    def remove_notes(self, notes: list[Note]) -> None:
        """Remove notes from the store.

        :param list[Note] notes: The notes to remove
        """
        for note in notes:
            for index in range(len(self.__list_store)):
                if note == self.__list_store[index]:
                    self.__list_store.remove(index)
                    break
            if note.id in self.__store_note_id_map:
                del self.__store_note_id_map[note.id]
            if note.has_remote_id and note.remote_id in self.__store_remote_id_map:
                del self.__store_remote_id_map[note.remote_id]

    def fetch_note_by_db_id(self, db_id: int) -> Optional[Note]:
        """Fetch a note by database id"

        :param int db_id: The database id of the note
        """
        if db_id not in self.__store_note_id_map:
            return None
        return self.__store_note_id_map[db_id]

    def ensure_remote_map_entry(self, note: Note) -> None:
        """Ensure the provided note has an entry in the remote id to note map.

        :param Note note: The note
        """
        if note.has_remote_id and note.remote_id not in self.__store_remote_id_map:
            self.__store_remote_id_map[note.remote_id] = note

    def fetch_note_by_remote_id(self, remote_id: int) -> Optional[Note]:
        """Fetch a note by remote id"

        :param int remote_id: The remote id of the note
        """
        if remote_id not in self.__store_remote_id_map:
            return None
        return self.__store_remote_id_map[remote_id]

    @staticmethod
    def __sort_func(note1: Note, note2: Note, _data: Any) -> int:
        """Sort notes.

        :param Note note1: First note
        :param Note note2: Second note
        :param _data: Unused
        :return: -1 if note1 earlier, 1 if later, 0 unused
        :rtype: int
        """
        if note1.favourite != note2.favourite:
            return -1 if note1.favourite else 1
        return -1 if note1.last_modified > note2.last_modified else 1


class NoteListModelCategoryFiltered(Gtk.FilterListModel):

    def __init__(self, model: Gio.ListModel) -> None:
        self.__filter = Gtk.CustomFilter()
        self.__filter.set_filter_func(self.__filter_func)
        self.__category: Optional[Category] = None
        super().__init__(model=model, filter=self.__filter)

    def invalidate_filter(self, category: Category) -> None:
        """Invalidate the filter.

        :param Category: Category to display
        """
        self.__category = category
        self.__filter.set_filter_func(self.__filter_func)

    def __filter_func(self, note: Note) -> bool:
        res = not note.locally_deleted
        if self.__category is not None:
            res = res and self.__category.includes_note(note)
        return res


class NoteListModelFavourites(Gtk.FilterListModel):

    def __init__(self, model: Gio.ListModel) -> None:
        self.__filter = Gtk.CustomFilter()
        self.invalidate_filter()
        super().__init__(model=model, filter=self.__filter)

    def invalidate_filter(self) -> None:
        """Invalidate the filter."""
        self.__filter.set_filter_func(NoteListModelFavourites.__filter_func)

    @staticmethod
    def __filter_func(note: Note) -> bool:
        """Filter for favourites.

        :param Note note: The note to filter
        """
        return note.favourite


class NoteListModelTimeFiltered(Gtk.FilterListModel):

    def __init__(self, model: Gio.ListModel) -> None:
        self.__filter = Gtk.CustomFilter()

        self.__time_min: Optional[int] = None
        self.__time_max: Optional[int] = None

        super().__init__(model=model, filter=self.__filter)

    def invalidate_filter(
        self, time_min: Optional[float] = None, time_max: Optional[float] = None
    ) -> None:
        """Invalidate the filter.

        :param Optional[float] time_min: Minimum date/time threshold timestamp
        :param Optional[float] time_max: Maximum date/time threshold timestamp
        """
        if time_min is not None:
            self.__time_min = int(time_min)
        else:
            self.__time_min = None
        if time_max is not None:
            self.__time_max = int(time_max)
        else:
            self.__time_max = None
        self.__filter.set_filter_func(self.__filter_func)

    def __filter_func(self, note: Note) -> bool:
        if note.favourite:
            include = False
        else:
            if self.__time_min is not None:
                include = note.last_modified > self.__time_min
            else:
                include = True
            if self.__time_max is not None:
                include = include and note.last_modified <= self.__time_max
        return include


class NoteListModelSearch(Gtk.FilterListModel):

    def __init__(self, model: Gio.ListModel) -> None:
        self.__filter = Gtk.CustomFilter()
        self.__filter.set_filter_func(self.__filter_func)

        self.__filter_to_ids: Optional[list[int]] = None

        super().__init__(model=model, filter=self.__filter)

    def invalidate(self, ids: Optional[list[int]]) -> None:
        """Invalidate filter, restricting by the provided note ids.

        :param Optional[list[int]] ids: The note ids (row primary key, not remote id) to be
            filtered by.
        """
        self.__filter_to_ids = ids
        self.__filter.set_filter_func(self.__filter_func)

    def __filter_func(self, note: Note) -> bool:
        if note.locally_deleted:
            include = False
        elif not self.__filter_to_ids:
            include = False
        else:
            include = note.id in self.__filter_to_ids
        return include



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod


class HtmlGeneratorConfiguration(ABC):
    """HTML generator configuration interface."""

    @property
    @abstractmethod
    def markdown_tex_support(self) -> bool:
        """Get whether markdown TeX rendering is supported.

        :return: Markdown TeX supported
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def line_length(self) -> int:
        """Get line length.

        :return: Size in pixels
        :rtype: int
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def markdown_use_monospace_font(self) -> bool:
        """Get whether to use a monospace font for the markdown render.

        :return: Using monospace font
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def use_monospace_font(self) -> bool:
        """Get whether to use a monospace font.

        :return: Using monospace font
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def font_size(self) -> int:
        """Get font size.

        :return: Size
        :rtype: int
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def high_contrast(self) -> bool:
        """Get whether high contrast output should be used for screen rendering.

        :return: Whether visible
        :rtype: bool
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def markdown_render_monospace_font_ratio(self) -> float:
        """Get the adjustment in size from proportional to fixed width font.

        :return: Ratio
        :rtype: float
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def editor_header_bar_visible_for_window_state(self) -> bool:
        """Get whether the header bar is configured visible for the window maximised state."

        :return: Whether visible
        :rtype: bool
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("GtkSource", "5")
from gi.repository import GLib, Gtk, GtkSource

import logging
import re
from typing import NamedTuple, Optional


class ParsedLink(NamedTuple):
    """Parsed general link."""

    link: str
    """Link URL"""

    start_offset: int
    """Start offset"""

    end_offset: int
    """End offset"""


class ParsedInlineLink(NamedTuple):
    """Parsed markdown inline link."""

    link: str = ""
    """Link URL"""

    text: str = ""
    """Link text"""

    source_text: str = ""
    """Source text"""

    start_offset: int = -1
    """Start offset"""

    end_offset: int = -1
    """End offset"""

    url_angle_bracketed: bool = False
    """Angle-bracket wrapped URL"""

    title: str = ""
    """Link title"""


class ParsedAutomaticLink(NamedTuple):
    """Parsed markdown automatic link."""

    link: str
    """Link URL"""

    start_offset: int
    """Start offset"""

    end_offset: int
    """End offset"""

    source_text: str
    """Source text"""


def sanitise_path(path: str) -> str:
    """Sanitise path for filesystem storage.

    :param str path: The path to work on
    :return: The sanitised path
    :rtype: str
    """
    # This is close to a direct translation of what Nextcloud Notes is using.

    # Remove characters which are illegal on Windows (includes illegal characters on Unix/Linux)
    # prevents also directory traversal by eliminiating slashes
    for token in ["*", "|", "/", "\\", ":", '"', "<", ">", "?"]:
        path = path.replace(token, "")

    # Prevent file being hidden
    path = re.sub(r"^[\.\s]+", "", path)

    return path.strip()


def has_unicode(input: str) -> bool:
    has_unicode = False
    try:
        input.encode(encoding="ascii")
    except UnicodeEncodeError:
        has_unicode = True
    return has_unicode


def parse_any_url_at_iter(iter: Gtk.TextIter, http_only: bool) -> Optional[ParsedLink]:
    """Check and parse location in buffer to URL.

    :param Gtk.TextIter iter: Iter
    :param bool http_only: Whether only http/s URIs will pass
    :return: Object with link details, optional
    :rtype: ParsedLink
    """
    start = iter.copy()
    end = iter.copy()

    def whitespace_at_iter(itr: Gtk.TextIter) -> bool:
        char_end = itr.copy()
        char_end.forward_char()
        char = itr.get_text(char_end)
        return char in (" ", "\t", "\n")

    # Find preceding whitespace
    moved = False
    while not start.is_start() and not whitespace_at_iter(start):
        moved = True
        start.backward_char()
    if moved and not (start.is_start() and not whitespace_at_iter(start)):
        start.forward_char()

    # Find following whitespace
    while not end.is_end() and not whitespace_at_iter(end):
        end.forward_char()

    # Grab text between linespace breaks and see if GLib thinks it's a URL, populate
    text = start.get_text(end)
    if str_is_url(text, http_only):
        return ParsedLink(text, start.get_offset(), end.get_offset())
    else:
        return None


def parse_markdown_inline_link(inside: Gtk.TextIter) -> Optional[ParsedInlineLink]:
    """Parse link elements from a known inline link.

    :param Gtk.TextIter inside: Iter inside inline link
    :return: Object with link details, optional
    :rtype: ParsedInlineLink
    """
    start, end = get_iters_at_sourceview_context_class_extents("inline-link", inside)
    if not start or not end:
        return None

    source_text = inside.get_buffer().get_text(start, end, include_hidden_chars=False)

    # Parsing logic here is a little brittle, reflecting being in this block is predicated on
    # the GtkSourceView language parsing
    text, url_part = source_text.split("](")
    text = text[1:]
    url_part = url_part[:-1].strip()
    title = ""
    if url_part.endswith('"'):
        # We have a title, parse that out
        url_part, title = url_part[:-1].split('"')
        url_part = url_part.strip()
    if url_part.startswith("<") and url_part.endswith(">"):
        link = url_part[1:-1]
        angle_bracketed = True
    else:
        link = url_part
        angle_bracketed = False

    return ParsedInlineLink(
        link, text, source_text, start.get_offset(), end.get_offset(), angle_bracketed, title
    )


def parse_markdown_automatic_link(inside: Gtk.TextIter) -> Optional[ParsedAutomaticLink]:
    """Parse URL from a known markdown automatic link.

    :param Gtk.TextIter inside: Iter inside inline link
    :return: Object with link details, optional
    :rtype: ParsedAutomaticLink
    """
    start, end = get_iters_at_sourceview_context_class_extents("automatic-link", inside)
    if not start or not end:
        return None
    link = start.get_text(end)
    if link.startswith("<") and link.endswith(">"):
        return ParsedAutomaticLink(link[1:-1], start.get_offset(), end.get_offset(), link)
    else:
        # Failure very unlikely as we depend on the GtkSourceView language parsing to reach
        # here
        logging.warning("Failed to parse automatic-link")
        return None


def get_iters_at_sourceview_context_class_extents(
    cls: str, inside: Gtk.TextIter
) -> tuple[Optional[Gtk.TextIter], Optional[Gtk.TextIter]]:
    """Get text iterators at the extents of a context class.

    :param str cls: Context class name
    :param Gtk.TextIter: Iterator inside class
    :return: The iters
    :rtype: tuple[Optional[Gtk.TextIter], Optional[Gtk.TextIter]]
    """
    start = inside.copy()
    if not iter_backward_to_context_class_start(start, cls):
        logging.warning("Backward to start of '{cls}' failed")
        return (None, None)
    end = start.copy()
    if not iter_forward_to_context_class_removed(end, cls):
        logging.warning("Forward to end of '{cls}' failed")
        return (None, None)
    return (start, end)


def iter_backward_to_context_class_start(location: Gtk.TextIter, cls: str) -> bool:
    """Move text iterator back to start of context class.

    :param Gtk.TextIter location: Iterator inside class
    :param str cls: Context class name
    :return: Success
    :rtype: bool
    """
    buffer: GtkSource.Buffer = location.get_buffer()
    if cls not in buffer.get_context_classes_at_iter(location):
        logging.debug(f"Not in '{cls}' at start to move backward")
        return False

    while not location.is_start():
        location.backward_char()
        if cls not in buffer.get_context_classes_at_iter(location):
            location.forward_char()
            return True
    return location.is_start()


def iter_forward_to_context_class_removed(location: Gtk.TextIter, cls: str) -> bool:
    """Move text iterator forward to end of context class.

    :param Gtk.TextIter location: Iterator inside class
    :param str cls: Context class name
    :return: Success
    :rtype: bool
    """
    buffer: GtkSource.Buffer = location.get_buffer()
    if cls not in buffer.get_context_classes_at_iter(location):
        logging.debug(f"Not in '{cls}' at start to move forward")
        return False

    while not location.is_end():
        location.forward_char()
        if cls not in buffer.get_context_classes_at_iter(location):
            return True
    return location.is_end()


def str_is_url(text: str, http_only: bool) -> bool:
    """Get whether provided text is a URL.

    :param str text: The text to check
    :param bool http_only: Whether only http/s URIs will pass
    :return: Whether a URL
    :rtype: bool
    """
    valid = False
    try:
        valid = GLib.uri_is_valid(text, GLib.UriFlags.NONE)
    except GLib.GError:
        pass
    # GLib's Uri.is_valid appears to allow eg. http:someurl.com
    if valid and http_only:
        scheme = GLib.uri_peek_scheme(text)
        if scheme not in ("http", "https") or not text.startswith(f"{scheme}://"):
            valid = False
    return valid



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib

from enum import IntEnum, auto
from io import BytesIO
import logging
import os
from pathlib import Path
from shutil import copyfile
from threading import Thread
from typing import BinaryIO, NamedTuple, Optional
from urllib.parse import quote

from markdown_it.token import Token

from iotas.attachment import Attachment
from iotas.markdown_helpers import (
    get_image_attachments_from_note_content,
    get_image_attachments_from_tokens,
    filter_image_tokens,
)
from iotas.note import Note

ATTACHMENT_DATA_SUBPATH = "iotas"
ATTACHMENT_DIR_NAME = "attachments"
ATTACHMENT_MAX_SIZE_MB = 50


class AttachmentDiskStates(NamedTuple):
    """Whether attachments exist on disk."""

    exists: dict[str, Attachment]
    """Found on disk"""

    missing: dict[str, Attachment]
    """Missing from disk"""


class AttachmentsCopyOutcome(IntEnum):
    """Outcome on attachments copy to disk path run."""

    NONE = auto()
    """There were no attachments to store"""

    FAILURE = auto()
    """One or more attachments failed to store"""

    HAD_MISSING = auto()
    """One or more attachments were missing on disk"""

    SUCCESS = auto()
    """All attachments were successfully copied"""


class AttachmentsCopyResults(NamedTuple):
    """Outcome and stored attachments for copy to disk path run."""

    outcome: AttachmentsCopyOutcome
    """Run outcome"""

    attachments: list[Attachment] = []
    """Attachments which were stored"""


def copy_note_attachments(note: Note, path: str, prefix_note_id: bool) -> AttachmentsCopyResults:
    """Copy note attachments to a directory.

    :param Note note: Note to work on
    :param str path: Destination directory
    :param bool prefix_note_id: Whether to prefix the note id
    :return: The outcome of the run and attachments which were stored
    :rtype: AttachmentsCopyResults
    """
    attachments = get_image_attachments_from_note_content(note)
    if not attachments:
        return AttachmentsCopyResults(AttachmentsCopyOutcome.NONE)

    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except OSError as e:
            logging.warning(f"Failed to create dir '{path}' to save attachments: {e}")
            return AttachmentsCopyResults(AttachmentsCopyOutcome.FAILURE)

    note_id_prefix = f"{note.id}." if prefix_note_id else ""
    had_failure = False
    had_missing = False
    stored = []

    for attachment in attachments:
        filename = f"{note_id_prefix}{os.path.basename(attachment.path)}"
        if not attachment_exists_locally(attachment):
            logging.warning(f"'{attachment.path}' not available to save")
            had_missing = True
            continue
        dest_path = os.path.join(path, filename)
        try:
            copyfile(get_attachment_filesystem_path(attachment), dest_path)
        except OSError as e:
            logging.warning(f"Failed to save attachment '{attachment.path}': {e}")
            had_failure = True
        else:
            stored.append(attachment)

    if had_failure:
        return AttachmentsCopyResults(AttachmentsCopyOutcome.FAILURE)
    elif had_missing:
        return AttachmentsCopyResults(AttachmentsCopyOutcome.HAD_MISSING, stored)
    else:
        return AttachmentsCopyResults(AttachmentsCopyOutcome.SUCCESS, stored)


def write_attachment_to_disk(attachment: Attachment, data: bytes) -> bool:
    """Write attachment to disk.

    Storing to user attachments directory.

    :param Attachment attachment: Attachment
    :param bytes data: Attachment data
    :return: Success
    :rtype: bool
    """
    path = get_attachment_filesystem_path(attachment)
    if os.path.exists(path):
        logging.warning(f"Attachment '{attachment.filename}' already exists for note?")
        return False

    try:
        with open(get_attachment_filesystem_path(attachment), "wb") as f:
            f.write(data)
    except OSError as e:
        logging.warning(f"Failed to write attachment to disk: {e}")
        return False
    else:
        return True


def get_attachments_dir() -> str:
    """Get user attachments directory.

    :return: Directory
    :rtype: str
    """
    return os.path.join(
        GLib.get_user_data_dir(),
        ATTACHMENT_DATA_SUBPATH,
        ATTACHMENT_DIR_NAME,
    )


def get_attachments_on_disk_for_note(note: Note) -> list[Attachment]:
    """Get the attachments on disk for a note.

    :param Note note: Note to match
    :return: Attachments
    :rtype: list[Attachment]
    """
    attachments = []
    path = Path(get_attachments_dir())
    prefix = f"{note.id}."

    # TODO Python 3.13+ remove exception handling
    try:
        files = list(path.glob(f"{prefix}*"))
    except OSError as e:
        logging.warning(f"Error listing attachments: {e}")
    else:
        for file in files:
            attachment = Attachment()
            attachment.note_id = note.id
            attachment.path = file.name[len(prefix) :]
            attachment.note_remote_id = note.remote_id
            attachments.append(attachment)

    return attachments


def delete_attachments_for_note(note: Note) -> None:
    """Delete attachments from disk for note

    :param Note note: Note to delete for
    """
    for attachment in get_attachments_on_disk_for_note(note):
        if not attachment_exists_locally(attachment):
            continue
        path = get_attachment_filesystem_path(attachment)
        try:
            os.remove(path)
        except OSError as e:
            logging.warning(f"Failed to delete attachment '{path}': {e}")


def get_attachment_disk_states(note: Note, tokens: list[Token]) -> AttachmentDiskStates:
    """Get whether a note's attachments are stored on disk.

    :param Note note: Note to check for
    :param Tokens tokens: Parser tokens to get list of attachments from
    :return: Whether on disk
    :rtype: AttachmentDiskStates
    """
    attachments = get_image_attachments_from_tokens(note, tokens)
    exists = {}
    missing = {}
    for attachment in attachments:
        if attachment_exists_locally(attachment):
            exists[attachment.path_quoted] = attachment
        else:
            missing[attachment.path_quoted] = attachment

    return AttachmentDiskStates(exists, missing)


def get_attachment_filesystem_uri(attachment: Attachment) -> str:
    """Get attachment filesystem path as URI.

    :param Attachment attachment: Attachment
    :return: Path as URI
    :rtype: str
    """
    return "file://" + quote(get_attachment_filesystem_path(attachment))


def get_attachment_filesystem_path(attachment: Attachment) -> str:
    """Get attachment filesystem path.

    :param Attachment attachment: Attachment
    :return: Path
    :rtype: str
    """
    filename = f"{attachment.note_id}.{attachment.filename}"
    return os.path.join(
        get_attachments_dir(),
        filename,
    )


def attachment_exists_locally(attachment: Attachment) -> bool:
    """Get whether attachment exists on disk.

    :param Attachment attachment: Attachment
    :return: Exists
    :rtype: bool
    """
    return os.path.exists(get_attachment_filesystem_path(attachment))


def get_attachment_file_object(attachment: Attachment) -> Optional[BinaryIO]:
    """Get attachment file object.

    :param Attachment attachment: Attachment
    :return: File object
    :rtype: BinaryIO, optional
    """
    content = None
    try:
        with open(get_attachment_filesystem_path(attachment), "rb") as f:
            content = f.read()
    except OSError as e:
        logging.warning(f"Failed to open attachment: {e}")
        return None

    return BytesIO(content)


def flush_all_attachments_on_disk() -> None:
    """Flush all attachments on disk."""
    for file in Path(get_attachments_dir()).iterdir():
        try:
            os.remove(file)
        except OSError as e:
            logging.warning(f"Error removing attachment '{file}': {e}")


def set_attachment_filesystem_uris_on_tokens(
    tokens: list[Token], new_paths: dict[str, Attachment]
) -> None:
    """Update image paths in parser tokens to filesystem URIs.

    :param list[Token] tokens: Parser tokens to update
    :param dict[str, Attachment] new_paths: Dict with (quoted) existing markup paths, attachments
        for filesystem URIs
    """
    images = filter_image_tokens(tokens)
    for img in images:
        src = img.attrs["src"]
        if src in new_paths:
            img.attrs["src"] = get_attachment_filesystem_uri(new_paths[src])


def trim_orphaned_images(note: Note, on_thread: bool) -> None:
    """Trim images on disk which don't appear in the note.

    :param Note note: Note to trim for
    :param bool on_thread: Whether to work on thread
    """

    def work(note: Note) -> None:
        on_disk_paths = [
            get_attachment_filesystem_path(attachment)
            for attachment in get_attachments_on_disk_for_note(note)
        ]
        if on_disk_paths:
            attachments = get_image_attachments_from_note_content(note)
            content_paths = [
                get_attachment_filesystem_path(attachment) for attachment in attachments
            ]
            for path in on_disk_paths:
                if path not in content_paths:
                    filename = os.path.basename(path)
                    logging.debug(
                        f"Deleting attachment no longer in note '{note.title}': '{filename}'"
                    )
                    try:
                        os.remove(path)
                    except OSError as e:
                        logging.warning(f"Error deleting attachment '{filename}': {e}")

    if on_thread:
        thread = Thread(target=work, args=(note,))
        thread.daemon = True
        thread.start()
    else:
        work(note)



"""Builds task/todo lists out of markdown lists with items starting with [ ] or [x]"""

# Ported by Wolmar Nyberg Åkerström from https://github.com/revin/markdown-it-task-lists
# ISC License
# Copyright (c) 2016, Revin Guillen
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Iotas notes:
#
# This was initially a hatchet job upon the original logic to allow for having
# elements eg. links and inline code blocks within the list items while
# retaining the ability to toggle the checkbox by clicking on the text.

# Issues with rendered line-wrapped output resulted in a further heavier
# hatchet job and stripping out of more of the original logic. Regardless,
# acknowledgements go to the original authors and porter for all the good
# parts of what's here.
#
# Original: https://github.com/executablebooks/mdit-py-plugins/

import re
from uuid import uuid4

from markdown_it import MarkdownIt
from markdown_it.rules_core.state_core import StateCore
from markdown_it.token import Token

# Regex string to match a whitespace character, as specified in
# https://github.github.com/gfm/#whitespace-character
# (spec version 0.29-gfm (2019-04-06))
_GFM_WHITESPACE_RE = r"[ \t\n\v\f\r]"


def tasklists_with_elements_plugin(
    md: MarkdownIt,
    enabled: bool = False,
):
    """Plugin for building task/todo lists out of markdown lists with items starting with [ ] or [x]
    .. Nothing else

    For example::
       - [ ] An item that needs doing
       - [x] An item that is complete

    The rendered HTML checkboxes are disabled; to change this, pass a truthy value into the enabled
    property of the plugin options.

    :param enabled: True enables the rendered checkboxes
    """
    disable_checkboxes = not enabled

    def fcn(state: StateCore) -> None:
        tokens: list[Token] = state.tokens

        i = 0
        while i < len(tokens):
            if is_todo_item(tokens, i):
                todoify(tokens[i], tokens[i].__class__)
                tokens[i - 2].attrSet(
                    "class",
                    "task-list-item" + (" enabled" if not disable_checkboxes else ""),
                )
                tokens[parent_token(tokens, i - 2)].attrSet("class", "contains-task-list")

                # Ensure any paragraph wrapping the li is hidden (as it breaks having elements
                # in the content)
                if not tokens[i - 1].hidden:
                    tokens[i - 1].hidden = True
            i += 1

    md.core.ruler.after("inline", "github-tasklists", fcn)

    def parent_token(tokens: list[Token], index: int) -> int:
        target_level = tokens[index].level - 1
        for i in range(1, index + 1):
            if tokens[index - i].level == target_level:
                return index - i
        return -1

    def is_todo_item(tokens: list[Token], index: int) -> bool:
        return (
            is_inline(tokens[index])
            and is_paragraph(tokens[index - 1])
            and is_list_item(tokens[index - 2])
            and starts_with_todo_markdown(tokens[index])
        )

    def todoify(token: Token, token_constructor: type) -> None:
        assert token.children is not None
        token.children.insert(0, make_checkbox(token, token_constructor))
        token.children[1].content = token.children[1].content[3:]
        token.content = token.content[3:]

        # Replaced number generator from original plugin with uuid.
        checklist_id = f"task-item-{uuid4()}"

        token.children[0].content = token.children[0].content[0:-1] + f' id="{checklist_id}"/>'
        token.children.insert(1, begin_div(token_constructor))
        token.children.append(end_div(token_constructor))

    def make_checkbox(token: Token, token_constructor: type) -> Token:
        checkbox = token_constructor("html_inline", "", 0)
        disabled_attr = 'disabled="disabled"' if disable_checkboxes else ""
        if token.content.startswith("[ ] "):
            checkbox.content = (
                f'<input class="task-list-item-checkbox" {disabled_attr} type="checkbox">'
            )
        elif token.content.startswith("[x] ") or token.content.startswith("[X] "):
            checkbox.content = (
                '<input class="task-list-item-checkbox" checked="checked" '
                f'{disabled_attr} type="checkbox">'
            )
        return checkbox

    def begin_div(token_constructor: type) -> Token:
        token = token_constructor("html_inline", "", 0)
        token.content = "<div>"
        return token

    def end_div(token_constructor: type) -> Token:
        token = token_constructor("html_inline", "", 0)
        token.content = "</div>"
        return token

    def is_inline(token: Token) -> bool:
        return token.type == "inline"

    def is_paragraph(token: Token) -> bool:
        return token.type == "paragraph_open"

    def is_list_item(token: Token) -> bool:
        return token.type == "list_item_open"

    def starts_with_todo_markdown(token: Token) -> bool:
        # leading whitespace in a list item is already trimmed off by markdown-it
        return re.match(rf"\[[ xX]]{_GFM_WHITESPACE_RE}+", token.content) is not None



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later
from gi.repository import Adw, Gdk, GLib, GObject, Gtk

from iotas.category import Category
from iotas.category_treeview_list_store import CategoryTreeViewListStore
from iotas.note import Note
from iotas.text_utils import sanitise_path
from iotas.ui_utils import check_for_open_first_result_shortcut


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/category_header_bar.ui")
class CategoryHeaderBar(Adw.Bin):
    __gtype_name__ = "CategoryHeaderBar"

    __gsignals__ = {
        "abort": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "categories-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    _entry: Gtk.Entry = Gtk.Template.Child()
    _completion: Gtk.EntryCompletion = Gtk.Template.Child()
    _apply_button: Gtk.Button = Gtk.Template.Child()
    _clear_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self.__model: CategoryTreeViewListStore
        self.__notes: list[Note] = []
        self.__changeset: list[tuple[Note, str]] = []

        completion_cell = Gtk.CellRendererText.new()
        completion_cell.set_property("xpad", 6)
        completion_cell.set_property("ypad", 6)
        self._completion.pack_start(completion_cell, False)
        self._completion.add_attribute(completion_cell, "text", 0)
        self._completion.connect("match-selected", self.__on_category_selected)
        self._entry.connect("icon-release", lambda _e, _p: self.__on_icon_released())
        self.__category_buffer = self._entry.get_buffer()
        self.__category_buffer.connect(
            "notify::text", lambda _o, _v: self.__update_clear_field_icon_visibility()
        )

    def setup(self, model: CategoryTreeViewListStore) -> None:
        """Perform initial setup.

        :param CategoryTreeViewListStore model: Categories model
        """
        self.__model = model
        self._completion.set_model(self.__model)
        self._completion.set_match_func(self.__completion_match_check)

    def replace_notes(self, notes: list[Note]) -> None:
        """Replace the notes on which the header is working.

        For selection changes in the index once the header is already visible.

        :param list[Note] notes: Notes to work on
        """
        self.__notes = notes.copy()
        self.__refresh_actions()

    def activate(self, notes: list[Note]) -> None:
        """Activate header bar.

        :param list[Note] notes: Notes to work on
        """
        self.__notes = notes.copy()
        if self.__model.iter_n_children(None) > 0:
            self.__show_popup()

        # If editing multiple notes only populate the entry if they're all the same
        uniform_category = notes[0].category
        if len(notes) > 1:
            for note in notes[1:]:
                if note.category != uniform_category:
                    uniform_category = ""
                    break
        self._entry.set_text(Category.generate_display_name(uniform_category))

        self._entry.grab_focus()
        self._entry.select_region(0, -1)
        self.__refresh_actions()

    def clear_and_apply(self) -> None:
        """Clear category and apply."""
        self._entry.set_text("")
        self.__apply_change()

    def take_changeset(self) -> list[tuple[Note, str]]:
        """Fetch changeset and clear.

        :return: List of modified notes with their previous categories
        :rtype: list[tuple[Note, str]]
        """
        changeset = self.__changeset
        self.__changeset = []
        return changeset

    def focus(self) -> None:
        """Grab focus to entry."""
        self._entry.grab_focus()

    def select_all_text(self) -> None:
        """Select all entry text"""
        self._entry.select_region(0, -1)

    @GObject.Property(type=bool, default=False)
    def entry_focused(self) -> bool:
        """Get whether the entry has focus."""
        entry_has_focus = self._entry.has_focus()
        # The Gtk.Text appears to get the focus, so check for that
        focused = self.get_root().get_focus()
        entry_child_has_focus = focused is not None and focused.is_ancestor(self._entry)
        return entry_has_focus or entry_child_has_focus

    @Gtk.Template.Callback()
    def _on_apply(self, _button: Gtk.Button) -> None:
        self.__apply_change()

    @Gtk.Template.Callback()
    def _on_entry_activated(self, _entry: Gtk.Entry) -> None:
        self.__apply_change()

    @Gtk.Template.Callback()
    def _on_clear(self, _button: Gtk.Button) -> None:
        self.clear_and_apply()

    @Gtk.Template.Callback()
    def _on_abort(self, _button: Gtk.Button) -> None:
        self.emit("abort")

    @Gtk.Template.Callback()
    def _on_entry_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if check_for_open_first_result_shortcut(keyval, state):
            self.__open_first_filter_match()

        return Gdk.EVENT_PROPAGATE

    def __on_icon_released(self) -> None:
        # Intentionally allowed for either icon
        if self._entry.get_text() == "":
            self.__show_popup()
        else:
            self._entry.get_buffer().delete_text(0, -1)
            self._entry.set_property("secondary-icon-name", "")

    def __on_category_selected(
        self, _completion: Gtk.EntryCompletion, store: Gtk.TreeModel, iter: Gtk.TreeIter
    ) -> bool:
        new_category = store.get_value(iter, 0)
        self.__update_category(new_category)
        return True

    def __apply_change(self) -> None:
        if not self.__notes:
            return
        new_category = self._entry.get_text().strip()
        self.__update_category(new_category)

    def __show_popup(self) -> None:
        child = self._entry.get_first_child()
        while not isinstance(child, Gtk.Popover):
            child = child.get_next_sibling()
        child.set_visible(True)
        self._completion.complete()

    def __refresh_actions(self) -> None:
        self._apply_button.set_sensitive(self.__notes)
        self._clear_button.set_sensitive(self.__notes)

    def __update_category(self, new_category: str) -> None:
        if "/" in new_category:
            parts = new_category.split("/")
            parts = [sanitise_path(x).strip() for x in parts]
            parts = [x for x in parts if x != ""]
            new_category = "/".join(parts)

        self.__changeset = []
        for note in self.__notes:
            if new_category != note.category:
                old_category = note.category
                note.category = new_category
                # Choosing to not bump last_modified for category updates
                note.flag_changed(update_last_modified=False)
                self.__changeset.append((note, old_category))
        GLib.idle_add(self.emit, "categories-changed")

    def __update_clear_field_icon_visibility(self) -> None:
        icon_showing = self._entry.get_property("secondary-icon_name")
        if self.__category_buffer.get_text() != "":
            if not icon_showing:
                self._entry.set_property("secondary-icon-name", "edit-clear-symbolic")
        else:
            if icon_showing:
                self._entry.set_property("secondary-icon-name", "")

    def __completion_match_check(
        self, _completion: Gtk.EntryCompletion, value: str, compare_iter: Gtk.TreeIter
    ) -> bool:
        """GtkEntryCompletion matching function.

        :param GtkEntryCompletion _completion: The EntryCompletion
        :param str value: The string to match
        :param GtkTreeIter compare_iter: The GtkTreeIter pointing at the value to compare
        :returns: Whether they match
        :rtype: bool
        """
        compare_category = self.__model.get_value(compare_iter, 1)
        if compare_category is None:
            # Seems to occur applying a dropdown value
            return False
        else:
            return self.__match_term_to_category(compare_category, value)

    def __match_term_to_category(self, category: Category, term: str) -> bool:
        return term.lower() in category.name.lower()

    def __open_first_filter_match(self) -> None:
        term = self._entry.get_text()
        for item in self.__model:
            category = self.__model.get_value(item.iter, 1)
            if self.__match_term_to_category(category, term):
                self._entry.set_text(category.name)
                self.__apply_change()
                break



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, GObject, Gtk

from iotas.text_utils import sanitise_path


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/editor_rename_header_bar.ui")
class EditorRenameHeaderBar(Adw.Bin):
    __gtype_name__ = "EditorRenameHeaderBar"
    __gsignals__ = {
        "cancelled": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "renamed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    _entry: Gtk.Entry = Gtk.Template.Child()
    _apply_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

    def enter(self, title: str) -> None:
        """Start renaming.

        :param str title: Existing title
        """
        self._entry.set_text(title)
        self._entry.grab_focus()
        self._entry.set_position(-1)

    def focus(self) -> None:
        """Grab focus to entry."""
        self._entry.grab_focus()

    @Gtk.Template.Callback()
    def _on_apply_pressed(self, _button: Gtk.Button) -> None:
        self.__apply_rename()

    @Gtk.Template.Callback()
    def _abort_rename(self, _button: Gtk.Button) -> None:
        self.emit("cancelled")

    @Gtk.Template.Callback()
    def _on_entry_changed(self, _entry: Gtk.Editable) -> None:
        self._apply_button.set_sensitive(self.__entry_valid())

    @Gtk.Template.Callback()
    def _on_entry_activated(self, _entry: Gtk.Entry) -> None:
        if self.__entry_valid():
            self.__apply_rename()

    def __apply_rename(self) -> None:
        self.emit("renamed", self.__get_sanitised_entry())

    def __entry_valid(self) -> bool:
        return self.__get_sanitised_entry() != ""

    def __get_sanitised_entry(self) -> str:
        new_title = self._entry.get_text().strip()
        return sanitise_path(new_title)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gi.repository import Gio, GLib

import itertools
import logging
import os
import sqlite3
from threading import current_thread, Lock
from typing import Optional

from iotas.config_manager import ConfigManager


class DbCursor:
    """
    Database cursor.
    """

    def __init__(self, obj: Database, commit: bool = False) -> None:
        self.__obj = obj
        self.__commit = commit
        self.__cursor: Optional[sqlite3.Connection] = None

    def __enter__(self) -> sqlite3.Connection:
        thread_id = current_thread().name
        app = Gio.Application.get_default()
        # App can be None for unit tests
        if app and thread_id in app.cursors.keys():
            cursor = app.cursors[thread_id]
            return cursor
        else:
            cursor = self.__obj.get_cursor()
            self.__cursor = cursor
            return cursor

    def __exit__(self, type, value, traceback) -> None:
        if self.__cursor is not None:
            if self.__commit:
                self.__obj.acquire_lock()
                self.__cursor.commit()
                self.__obj.release_lock()
            self.__cursor.close()
        self.__cursor = None


class DbLock:
    """
    Database lock.
    """

    def __init__(self) -> None:
        self.__lock = Lock()
        self.__locked = False

    def acquire(self) -> None:
        """Acquire lock."""
        self.__locked = True
        self.__lock.acquire()

    def release(self) -> None:
        """Release lock."""
        self.__locked = False
        self.__lock.release()

    @property
    def locked(self) -> bool:
        return self.__locked


class Database:
    """
    Base database object.
    """

    DATA_SUBPATH = "iotas"
    FILENAME = "notes.db"
    VERSION = 3

    __fts_creation = """
    CREATE VIRTUAL TABLE note_fts_idx
    USING fts5(title, note_content, content='note', content_rowid='id');

    CREATE TRIGGER note_ai AFTER INSERT ON note BEGIN
      INSERT INTO note_fts_idx(rowid, title, note_content)
      VALUES (new.id, new.title, new.content);
    END;
    CREATE TRIGGER note_ad AFTER DELETE ON note BEGIN
      INSERT INTO note_fts_idx(note_fts_idx, rowid, title, note_content)
      VALUES('delete', old.id, old.title, old.content);
    END;
    CREATE TRIGGER note_au AFTER UPDATE ON note BEGIN
      INSERT INTO note_fts_idx(note_fts_idx, rowid, title, note_content)
      VALUES('delete', old.id, old.title, old.content);
      INSERT INTO note_fts_idx(rowid, title, note_content)
      VALUES (new.id, new.title, new.content);
    END;
    """

    __create_script = f"""
    CREATE TABLE note (id INTEGER PRIMARY KEY,
                       title TEXT NOT NULL,
                       content TEXT NOT NULL,
                       excerpt TEXT NOT NULL,
                       last_modified INT NOT NULL,
                       favourite BOOLEAN NOT NULL,
                       category TEXT NOT NULL,
                       remote_id INT NOT NULL,
                       dirty BOOLEAN NOT NULL,
                       locally_deleted BOOLEAN NOT NULL,
                       etag TEXT,
                       read_only BOOLEAN NOT NULL DEFAULT false,
                       failed_pushes TEXT);
    CREATE index idx_note_last_modified ON note(last_modified);
    CREATE index idx_note_remote_id ON note(remote_id);

    {__fts_creation}
    """

    # A migration is a tuple comprising the query and whether we want to resync from the server
    __migrations = {}
    __migrations[1] = ("ALTER TABLE note ADD COLUMN read_only BOOLEAN NOT NULL DEFAULT false", True)
    __second_migration = f"""{__fts_creation}
    INSERT INTO note_fts_idx (rowid, title, note_content)
    SELECT id, title, content FROM note;
    """
    __migrations[2] = (__second_migration, True)
    __migrations[3] = ("ALTER TABLE note ADD COLUMN failed_pushes TEXT", False)

    def __init__(self, custom_directory: Optional[str] = None) -> None:
        self.__thread_lock = DbLock()

        # Allow for custom directory for testing
        if custom_directory:
            self.__db_directory = custom_directory
        else:
            self.__db_directory = os.path.join(GLib.get_user_data_dir(), self.DATA_SUBPATH)
        self.__db_path = os.path.join(self.__db_directory, self.FILENAME)

        f = Gio.File.new_for_path(self.__db_path)
        if not f.query_exists():
            try:
                d = Gio.File.new_for_path(self.__db_directory)
                if not d.query_exists():
                    d.make_directory_with_parents()
                with DbCursor(self, True) as sql:
                    sql.executescript(self.__create_script)
                    sql.execute("PRAGMA user_version=%s" % self.VERSION)
            except Exception as e:
                logging.error("DB init error: {}".format(e))
                raise (e)
        else:
            current_version = self.__get_version()
            if current_version < self.VERSION:
                self.__run_migrations(current_version)
            elif current_version > self.VERSION:
                versions = f"v{current_version} > v{self.VERSION}"
                logging.warning(f"Database schema version is newer than expected ({versions})")

    def acquire_lock(self):
        self.__thread_lock.acquire()

    def release_lock(self):
        self.__thread_lock.release()

    def execute(self, request: str) -> list:
        """Execute a query

        :param str request: The request
        :return: A list of ids, if applicable
        :rtype: list
        """
        try:
            ids = []
            db = Gio.Application.get_default().db
            with DbCursor(db) as sql:
                result = sql.execute(request)
                ids += list(itertools.chain(*result))
                return ids
        except Exception as e:
            logging.error("DB execute error: {} for {}".format(e, request))
        return []

    def get_cursor(self) -> sqlite3.Connection:
        """Get a cursor.

        :return: The cursor
        :rtype: sqlite3.Connection
        """
        return sqlite3.connect(self.__db_path)

    def trash(self) -> None:
        """Trash the database, remove the file."""
        os.remove(self.__db_path)

    def __get_version(self) -> int:
        request = "PRAGMA user_version"
        try:
            version = None
            with DbCursor(self) as sql:
                result = sql.execute(request)
                row = result.fetchone()
                if isinstance(row, tuple) and len(row) == 1:
                    version = row[0]
        except Exception as e:
            logging.error("DB execute error: {} for {}".format(e, request))
        if not version:
            logging.warning("Didn't get version from DB, using 1")
            version = 1
        return version

    def __set_version(self, version: int) -> None:
        if not isinstance(version, int):
            return
        request = f"PRAGMA user_version = {version}"
        try:
            with DbCursor(self) as sql:
                sql.execute(request)
        except Exception as e:
            logging.error("DB execute error: {} for {}".format(e, request))

    def __run_migrations(self, current_version: int) -> None:
        logging.info(
            "Running database schema migration from v{} to v{}".format(
                current_version, self.VERSION
            )
        )
        resync_requested = False
        in_error = False
        while current_version != self.VERSION:
            request, resync = self.__migrations[current_version + 1]
            try:
                with DbCursor(self) as sql:
                    sql.executescript(request)
            except Exception as e:
                logging.error("DB execute error: {} for {}".format(e, request))
                in_error = True
                break
            current_version += 1
            if resync:
                resync_requested = True

        if resync_requested:
            ConfigManager.get_default().nextcloud_prune_threshold = 0

        if not in_error:
            self.__set_version(current_version)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod


class NextcloudSyncWorkerConfiguration(ABC):
    """Nextcloud sync worker configuration interface."""

    @property
    @abstractmethod
    def nextcloud_endpoint(self) -> str:
        """Get Nextcloud endpoint.

        :return: Endpoint
        :rtype: str
        """
        raise NotImplementedError()

    @nextcloud_endpoint.setter
    @abstractmethod
    def nextcloud_endpoint(self, value: str) -> None:
        """Set Nextcloud endpoint.

        :param str value: New value
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def nextcloud_username(self) -> str:
        """Get Nextcloud username.

        :return: Username
        :rtype: str
        """
        raise NotImplementedError()

    @nextcloud_username.setter
    @abstractmethod
    def nextcloud_username(self, value: str) -> None:
        """Set Nextcloud username.

        :param str value: New value
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def nextcloud_prune_threshold(self) -> int:
        """Get Nextcloud prune threshold.

        :return: Threshold
        :rtype: int
        """
        raise NotImplementedError()

    @nextcloud_prune_threshold.setter
    @abstractmethod
    def nextcloud_prune_threshold(self, value: int) -> None:
        """Set Nextcloud prune threshold.

        :param int value: New value
        """
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import warnings

from gi.repository import GtkSource

from iotas.list_formatter import check_and_extend_list, increase_indentation, decrease_indentation

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    buffer: GtkSource.Buffer

    def test_check_and_extend_list(self) -> None:
        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "\n")
        check_and_extend_list(buffer)
        self.assertEqual(self.__get_text(), "\n")

        for char in "-+*":
            buffer = self.__reset()
            buffer.insert(buffer.get_start_iter(), f"{char} test")
            check_and_extend_list(buffer)
            self.assertEqual(self.__get_text(), f"{char} test\n{char} ")

            buffer = self.__reset()
            buffer.insert(buffer.get_start_iter(), f"  {char} test")
            check_and_extend_list(buffer)
            self.assertEqual(self.__get_text(), f"  {char} test\n  {char} ")

        for separator in ".)":
            buffer = self.__reset()
            buffer.insert(buffer.get_start_iter(), f"1{separator} test")
            check_and_extend_list(buffer)
            self.assertEqual(self.__get_text(), f"1{separator} test\n2{separator} ")

            buffer = self.__reset()
            buffer.insert(buffer.get_start_iter(), f"  1{separator} test")
            check_and_extend_list(buffer)
            self.assertEqual(self.__get_text(), f"  1{separator} test\n  2{separator} ")

            buffer = self.__reset()
            buffer.insert(buffer.get_start_iter(), f"a{separator} test")
            check_and_extend_list(buffer)
            self.assertEqual(self.__get_text(), f"a{separator} test\nb{separator} ")

            buffer = self.__reset()
            buffer.insert(buffer.get_start_iter(), f"  a{separator} test")
            check_and_extend_list(buffer)
            self.assertEqual(self.__get_text(), f"  a{separator} test\n  b{separator} ")

    def test_decrease_indentation(self) -> None:
        buffer = self.__reset()
        decrease_indentation(buffer)
        self.assertEqual(self.__get_text(), "")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "\tplain")
        decrease_indentation(buffer)
        self.assertEqual(self.__get_text(), "plain")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "  - text")
        decrease_indentation(buffer)
        self.assertEqual(self.__get_text(), "- text")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "  - text\n  - text\n  - text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        decrease_indentation(buffer)
        self.assertEqual(self.__get_text(), "- text\n- text\n- text")

    def test_increase_indentation(self) -> None:
        buffer = self.__reset()
        increase_indentation(buffer)
        self.assertEqual(self.__get_text(), "\t")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "plain")
        increase_indentation(buffer)
        self.assertEqual(self.__get_text(), "plain\t")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "plain")
        buffer.place_cursor(self.buffer.get_start_iter())
        increase_indentation(buffer)
        self.assertEqual(self.__get_text(), "\tplain")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "- text")
        increase_indentation(buffer)
        self.assertEqual(self.__get_text(), "  - text")

        buffer = self.__reset()
        buffer.insert(buffer.get_start_iter(), "- text\n- text\n- text")
        buffer.select_range(buffer.get_start_iter(), buffer.get_end_iter())
        increase_indentation(buffer)
        self.assertEqual(self.__get_text(), "  - text\n  - text\n  - text")

    def __get_text(self) -> str:
        if self.buffer is None:
            return ""
        else:
            start = self.buffer.get_start_iter()
            end = self.buffer.get_end_iter()
            return self.buffer.get_text(start, end, False)

    def __reset(self) -> GtkSource.Buffer:
        self.buffer = GtkSource.Buffer.new()
        return self.buffer



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, Gtk

import logging

from iotas.config_manager import ConfigManager


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/font_size_selector.ui")
class FontSizeSelector(Gtk.Box):
    __gtype_name__ = "FontSizeSelector"

    _label: Gtk.Label = Gtk.Template.Child()

    VALID_SIZES = [
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        20,
        22,
        24,
        26,
        28,
        30,
        34,
        38,
    ]

    def __init__(self) -> None:
        super().__init__()
        self.__increase_action: Gio.SimpleAction
        self.__decrease_action: Gio.SimpleAction
        self.__config_manager = ConfigManager.get_default()
        self.__config_manager.connect_changed(
            ConfigManager.FONT_SIZE,
            self.__refresh_from_setting,
        )

    def setup(self):
        self.__setup_actions()
        self.__refresh_from_setting()

    def increase(self) -> None:
        if not self.__increase_action.get_enabled():
            return

        previous_size = self.__config_manager.font_size
        ind = self.VALID_SIZES.index(previous_size)
        size = self.VALID_SIZES[ind + 1]
        self.__config_manager.font_size = size
        logging.info(f"Font size increased to {size}pt")

    def decrease(self) -> None:
        if not self.__decrease_action.get_enabled():
            return

        previous_size = self.__config_manager.font_size
        ind = self.VALID_SIZES.index(previous_size)
        size = self.VALID_SIZES[ind - 1]
        self.__config_manager.font_size = size
        logging.info(f"Font size decreased to {size}pt")

    def reset(self) -> None:
        previous_size = self.__config_manager.font_size
        default_size = self.__config_manager.default_font_size
        if previous_size != default_size:
            self.__config_manager.font_size = default_size
            logging.info(f"Font size reset (to {default_size}pt)")

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("increase")
        action.connect("activate", lambda _o, _v: self.increase())
        action_group.add_action(action)
        self.__increase_action = action

        action = Gio.SimpleAction.new("decrease")
        action.connect("activate", lambda _o, _v: self.decrease())
        action_group.add_action(action)
        self.__decrease_action = action

        action = Gio.SimpleAction.new("reset")
        action.connect("activate", lambda _o, _v: self.reset())
        action_group.add_action(action)

        app.get_active_window().insert_action_group("font-size-selector", action_group)
        self.__action_group = action_group

    def __refresh_from_setting(self) -> None:
        size = self.__config_manager.font_size
        self._label.set_label("{}pt".format(size))
        ind = self.VALID_SIZES.index(size)
        self.__increase_action.set_enabled(ind + 1 < len(self.VALID_SIZES))
        self.__decrease_action.set_enabled(ind - 1 >= 0)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GLib, GObject, Gtk

import logging
from typing import Optional

from iotas.attachment import Attachment
from iotas.attachment_helpers import trim_orphaned_images
from iotas.config_manager import ConfigManager
from iotas.category_manager import CategoryManager
import iotas.const as const
from iotas.editor import Editor, EditorSessionDetails
from iotas.export_dialog import ExportDialog
from iotas.index import Index
from iotas.note import Note
from iotas.note_database import NoteDatabase
from iotas.note_manager import NoteManager
from iotas.nextcloud_login_dialog import NextcloudLoginDialog, LoginDialogMode
from iotas.sync_manager import SyncManager
from iotas.uri_to_note_creator import UriToNoteCreator


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/window.ui")
class Window(Adw.ApplicationWindow):
    __gtype_name__ = "Window"

    _navigation: Adw.NavigationView = Gtk.Template.Child()
    _index: Index = Gtk.Template.Child()
    _editor: Editor = Gtk.Template.Child()
    _editor_page: Adw.NavigationPage = Gtk.Template.Child()

    def __init__(
        self,
        app: Adw.Application,
        db: NoteDatabase,
        sync_manager: SyncManager,
    ) -> None:
        super().__init__(application=app)

        self.set_icon_name(const.APP_ID)

        if const.IS_DEVEL:
            self.add_css_class("devel")

        self.app = self.get_application()

        self.__db = db
        self.__sync_manager = sync_manager

        self.__category_manager = CategoryManager(self.__db)
        self.__note_manager = NoteManager(self.__db, self.__category_manager)
        self.__config_manager = ConfigManager.get_default()

        self.__initialised = False
        self.__using_keyboard_navigation = False
        self.__secret_service_failed = False
        self.__open_note_after_init = None
        self.__create_note_after_init = False

        self.__setup_actions()

        self._index.setup(self.__note_manager, self.__category_manager, self.__sync_manager)
        self._index.connect(
            "note-opened", lambda _o, note, immediate: self.__open_note(note, immediate)
        )
        self._index.connect("reauthenticate", lambda _o: self.show_login_dialog())
        self._index.connect("export", lambda _o: self.__export_notes())
        self._index.connect(
            "header-bar-height", lambda _o, height: self._editor.set_header_bar_height(height)
        )

        self._editor.setup(self.__category_manager.tree_store, self.__sync_manager)
        self._editor.connect(
            "note-modified",
            lambda _o, note: self.__note_manager.persist_note_while_editing(
                note, update_excerpt=False
            ),
        )
        self._editor.connect("note-deleted", self.__on_editor_note_deleted)
        self._editor.connect(
            "category-changed",
            lambda _o, note, old_category: self.__note_manager.persist_note_category(
                note, old_category
            ),
        )
        self._editor.connect("exit", lambda _o: self._navigation.pop())

        self.__note_manager.connect("sync-requested", lambda _o: self.__sync_manager.sync_now())

        self.__sync_manager.connect("ready", self.__on_sync_ready)
        self.__sync_manager.connect("note-conflict", lambda _o, note: self.__on_note_conflict(note))
        self.__sync_manager.connect(
            "note-deleted", lambda _o, note: self.__on_note_remotely_deleted(note)
        )
        self.__sync_manager.connect("secret-service-failure", self.__on_secret_service_failure)
        self.__sync_manager.connect(
            "notify::authenticated",
            lambda _o, _v: self._editor.set_sync_authenticated(self.__sync_manager.authenticated),
        )
        self.__sync_manager.connect(
            "remote-category-sanitised",
            lambda _o, old, new: self.__category_manager.note_category_changed(old, new),
        )
        self.__sync_manager.connect("attachment-downloaded", self.__on_attachment_downloaded)

        self.__note_manager.connect(
            "initial-load-complete", lambda _o: self.__on_initial_load_from_db_complete()
        )

        self.connect("close-request", lambda _o: self.cleanup_and_close())
        self.connect("notify::visible-dialog", lambda _o, _v: self.__update_for_visible_dialogs())

        width, height = self.__config_manager.get_window_size()
        self.set_default_size(width, height)
        self._index.update_layout_for_initial_size(width, height)
        self.connect(
            "notify::default-height",
            lambda _o, _v: self._index.update_search_pagesize_for_height(self.get_height()),
        )

        self._index.active = True
        self._editor.active = False

    def cleanup_and_close(self) -> None:
        """Cleanup for exit."""
        if self._editor.editing:
            self.__note_manager.apply_any_server_sanitised_title(self._editor.current_note)
            self.__note_manager.persist_note_sync(
                self._editor.current_note, force_update_excerpt=self._editor.note_had_edit
            )
            trim_orphaned_images(self._editor.current_note, on_thread=False)

        self.__note_manager.apply_deleted_notes_attachment_deletion()
        self.__sync_manager.close()
        if not self.is_fullscreen():
            self.__config_manager.set_window_size(self.get_width(), self.get_height())
        self.__config_manager.first_start = False
        self.close()

    def show_login_dialog(self) -> None:
        """Show Nextcloud Notes onboarding dialog."""
        if self.__secret_service_failed:
            mode = LoginDialogMode.SECRET_SERVICE_FAILURE
        elif self.__sync_manager.configured_but_no_password:
            mode = LoginDialogMode.REAUTHENTICATE
        else:
            mode = LoginDialogMode.NORMAL
        dialog = NextcloudLoginDialog(self.__sync_manager, mode)
        dialog.present(self)

    @GObject.Property(type=bool, default=False)
    def using_keyboard_navigation(self) -> bool:
        return self.__using_keyboard_navigation

    @using_keyboard_navigation.setter
    def set_using_keyboard_navigation(self, value: bool) -> None:
        self.__using_keyboard_navigation = value

    @Gtk.Template.Callback()
    def _on_page_pushed(self, _obj: GObject.Object) -> None:
        self._index.active = False
        self._editor.active = True
        self._editor.focus_textview_if_editing()

    @Gtk.Template.Callback()
    def _on_page_popped(self, _obj: GObject.Object, _page: Adw.NavigationPage) -> None:
        note = self._editor.current_note
        self.__note_manager.persist_note_sync(note, force_update_excerpt=self._editor.note_had_edit)
        self._index.refresh_after_note_closed(note)
        self._editor.close_note()
        on_thread = True
        GLib.timeout_add(500, trim_orphaned_images, note, on_thread)
        if note.dirty:
            self.__sync_manager.sync_now()
        self._index.active = True
        self._editor.active = False

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("fullscreen")
        action.connect("activate", lambda _a, _p: self.__toggle_fullscreen())
        action_group.add_action(action)
        app.set_accels_for_action("win.fullscreen", ["F11"])

        action = Gio.SimpleAction.new("close")
        action.connect("activate", lambda _o, _v: self.cleanup_and_close())
        action_group.add_action(action)
        app.set_accels_for_action("win.close", ["<Control>w"])
        self.__close_window_action = action

        action = Gio.SimpleAction.new("refresh")
        action.connect("activate", lambda _o, _v: self.__sync_manager.sync_now())
        action_group.add_action(action)
        app.set_accels_for_action("win.refresh", ["<Control>r"])
        action.set_enabled(False)

        action = Gio.SimpleAction.new("start-nextcloud-signin")
        action.connect("activate", lambda _o, _v: self.show_login_dialog())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("create-note-from-cli")
        action.connect("activate", lambda _o, _v: self.__on_create_note())
        action_group.add_action(action)

        action = Gio.SimpleAction.new("create-note-from-uri", GLib.VariantType("s"))
        action.connect("activate", self.__on_create_note_from_uri)
        action_group.add_action(action)

        action = Gio.SimpleAction.new("open-note", GLib.VariantType("u"))
        action.connect("activate", self.__on_open_note)
        action_group.add_action(action)

        action = Gio.SimpleAction.new("open-previous-note")
        action.connect("activate", lambda _o, _v: self.__open_previous_note())
        app.set_accels_for_action("win.open-previous-note", ["<Control>l"])
        action_group.add_action(action)

        action = Gio.SimpleAction.new("search-from-cli", GLib.VariantType("s"))
        action.connect("activate", self.__on_search_from_cli)
        action_group.add_action(action)

        self.insert_action_group("win", action_group)
        self.__action_group = action_group

    def __on_editor_note_deleted(self, _obj: GObject.Object, note: Note) -> None:
        if self._navigation.get_visible_page() is not self._editor_page:
            return

        self._navigation.pop()
        self._index.delete_notes([note])

    def __on_note_conflict(self, note: Note) -> None:
        if self._editor.current_note is not None and note is self._editor.current_note:
            logging.info("Note being edited in conflict")
            self._editor.current_note.handling_conflict = True
            self._navigation.pop()
            self._index.show_note_conflict_alert()

    def __on_note_remotely_deleted(self, note: Note) -> None:
        if self._editor.current_note is not None and note is self._editor.current_note:
            logging.info("Note being edited was remotely deleted")
            self._editor.current_note.handling_conflict = True
            self._navigation.pop()
            self._index.show_editing_note_deleted_alert()

    def __on_sync_ready(self, _obj: GObject.Object, new_setup: bool) -> None:
        action = self.__action_group.lookup("refresh")
        action.set_enabled(True)
        action = self.__action_group.lookup("start-nextcloud-signin")
        action.set_enabled(False)

    def __toggle_fullscreen(self) -> None:
        if self.is_fullscreen():
            self.unfullscreen()
        else:
            self.fullscreen()

    def __on_secret_service_failure(self, _obj: GObject.Object) -> None:
        self.__secret_service_failed = True
        if self.__config_manager.show_startup_secret_service_failure:
            self._index.show_secret_service_failure_alert()

    def __on_create_note(self) -> None:
        """Create a new note and edit it."""
        if self.__initialised:
            if self._editor.active:
                self._navigation.pop()
            self.activate_action("index.create-note")
        else:
            self.__create_note_after_init = True

    def __on_open_note(
        self,
        _obj: GObject.Object,
        note_id: GObject.ParamSpec,
    ) -> None:
        note_id = note_id.get_uint32()
        if self.__initialised:
            if self._editor.active:
                self._navigation.pop()
            self.__open_note_by_id(note_id, immediate=False)
        else:
            self.__open_note_after_init = note_id

    # Note: Iotas' addition of this URI handler is a stopgap measure until the freedesktop
    # intent system is fleshed out. The handling provided here will be unceremoniously removed
    # when that system is available and may be removed or modified without notice beforehand;
    # this is not a stable public interface.
    def __on_create_note_from_uri(
        self,
        _obj: GObject.Object,
        param: GObject.ParamSpec,
    ) -> None:
        uri = param.get_string()

        parser = UriToNoteCreator()
        result = parser.parse(uri)
        if result.success:
            assert result.note
            note = result.note
            logging.info(f"Creating note from URI '{note.title}'")
            self.__note_manager.persist_note_sync(note, force_update_excerpt=False)
            self.__sync_manager.sync_now()

            if result.open:
                if self._editor.active:
                    self._navigation.pop()
                self.__open_note(note, immediate=True)

    def __on_search_from_cli(
        self,
        _obj: GObject.Object,
        param: GObject.ParamSpec,
    ) -> None:
        search_term = param.get_string()
        if self._editor.active:
            self._navigation.pop()
        self._index.search_from_cli(search_term)

    def __on_initial_load_from_db_complete(self) -> None:
        self.__initialised = True
        if self.__create_note_after_init:
            note = self.__note_manager.create_note(self.__category_manager.all_category)
            self.__open_note(note, immediate=True)
        elif self.__open_note_after_init is not None:
            self.__open_note_by_id(self.__open_note_after_init, immediate=True)

    def __on_attachment_downloaded(self, _obj: GObject.Object, attachment: Attachment) -> None:
        if not self._editor.active or not self._editor.editing:
            return
        self._editor.show_downloaded_attachment_in_render(attachment)

    def __open_note(
        self, note: Note, immediate: bool, restore_session: Optional[EditorSessionDetails] = None
    ) -> None:
        if immediate:
            self._navigation.set_animate_transitions(False)
        if not note.content_loaded:
            self.__db.populate_note_content(note)
        # Handle opening a note from CLI / shell search provider when another note is open
        if self._editor.active:
            self._navigation.pop()
        self._editor.init_note(note, restore_session)
        self._navigation.push(self._editor_page)
        if immediate:
            self._navigation.set_animate_transitions(True)

    def __open_note_by_id(self, note_id: int, immediate: bool) -> None:
        note = self.__note_manager.fetch_note_by_id(note_id)
        if note is None:
            logging.error(f"Failed to find note with id {note_id}")
            return
        self.__open_note(note, immediate)

    def __open_previous_note(self) -> None:
        sessions = self._editor.get_previous_sessions()
        for session in reversed(sessions):
            if session is None:
                continue
            if not self.__note_manager.fetch_note_by_id(session.note.id):
                continue
            if session.note.locally_deleted:
                continue
            if session.note in self.__note_manager.get_deleted_notes_pending_undo():
                continue
            if self._editor.current_note == session.note:
                continue

            logging.info("Opening previous note")
            session.direct_switching = self._editor.active
            self.__open_note(session.note, immediate=self._editor.active, restore_session=session)
            break

    def __update_for_visible_dialogs(self) -> None:
        visible = self.get_visible_dialog() is not None
        self._index.update_for_dialog_visibility(visible)
        self._editor.update_for_dialog_visibility(visible)
        self.__close_window_action.set_enabled(not visible)

    def __export_notes(self) -> None:
        html_generator = self._editor.ensure_html_generator_initialised()

        notes = self._index.get_selection()
        if not notes:
            return

        dialog = ExportDialog(
            notes,
            html_generator,
            self.__sync_manager.download_attachment if self.__sync_manager.authenticated else None,
            self._editor.ensure_webkit_initialised,
            self.__db.populate_note_content,
        )
        dialog.present(self)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import unittest
import warnings

from iotas.ordered_list_utils import (
    check_line_for_ordered_list_item,
    calculate_ordered_list_index,
    format_ordered_list_item,
)

warnings.filterwarnings("ignore", "version")


class Test(unittest.TestCase):

    def test_check_line_for_ordered_list_item(self) -> None:
        self.assertIsNotNone(check_line_for_ordered_list_item("1. item"))
        self.assertIsNotNone(check_line_for_ordered_list_item("1) item"))
        self.assertIsNotNone(check_line_for_ordered_list_item("a) item"))
        self.assertIsNotNone(check_line_for_ordered_list_item("a. item"))
        self.assertIsNone(check_line_for_ordered_list_item("plain"))
        self.assertIsNone(check_line_for_ordered_list_item(""))
        self.assertIsNone(check_line_for_ordered_list_item("1 item"))
        self.assertIsNone(check_line_for_ordered_list_item("- unordered"))

    def test_calculate_ordered_list_index(self) -> None:
        self.assertEqual(calculate_ordered_list_index("1", 1), "2")
        self.assertEqual(calculate_ordered_list_index("2", -1), "1")
        self.assertEqual(calculate_ordered_list_index("1", -1), "0")
        self.assertEqual(calculate_ordered_list_index("0", -1), "-1")

        self.assertEqual(calculate_ordered_list_index("a", 1), "b")
        self.assertEqual(calculate_ordered_list_index("A", 1), "B")
        self.assertEqual(calculate_ordered_list_index("b", -1), "a")
        self.assertEqual(calculate_ordered_list_index("B", -1), "A")
        self.assertIsNone(calculate_ordered_list_index("z", 1))
        self.assertIsNone(calculate_ordered_list_index("Z", 1))

        self.assertIsNone(calculate_ordered_list_index("1.", 1))
        self.assertIsNone(calculate_ordered_list_index("a)", 1))
        self.assertIsNone(calculate_ordered_list_index("", 1))
        self.assertIsNone(calculate_ordered_list_index("text", 1))

    def test_format_ordered_list_item(self) -> None:
        self.assertEqual(format_ordered_list_item("", "a", ")"), "a) ")
        self.assertEqual(format_ordered_list_item("  ", "3", "."), "  3. ")



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

import gi

gi.require_version("GtkSource", "5")
from gi.repository import Gdk, Gio, GLib, GObject, Gtk, GtkSource

from gtkspellcheck import SpellChecker, NoDictionariesFound

import locale
import logging
from typing import Optional

from iotas.config_manager import ConfigManager
from iotas import list_formatter


class EditorTextView(GtkSource.View):
    __gtype_name__ = "EditorTextView"

    __gsignals__ = {
        "margins-updated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    SOURCEVIEW_NO_SPELLCHECK_TAG = "gtksourceview:context-classes:no-spell-check"
    LINK_TAGS = {"inline-link", "automatic-link", "bare-link"}

    STANDARD_X_MARGIN = 36
    MINIMUM_MARGIN = 10

    def __init__(self) -> None:
        super().__init__()

        self.__css_provider: Optional[Gtk.CssProvider] = None
        self.__config_manager = ConfigManager.get_default()
        self.__line_length = -1
        self.__header_bar_height = 0
        self.__longpress_popover: Gtk.PopoverMenu
        self.__font_family = "monospace"
        self.__width_margins_set_for = None
        self.__height_margins_set_for: Optional[int] = None

        self.__cursor_pointer = Gdk.Cursor.new_from_name("pointer")
        self.__showing_pointer_cursor = False
        self.__control_down = False
        self.__mouse_x: Optional[int] = None
        self.__mouse_y: Optional[int] = None

        self.__config_manager.connect_changed(ConfigManager.FONT_SIZE, self.__update_font)

        self.__config_manager.connect_changed(
            ConfigManager.EDITOR_HEADER_BAR_VISIBILTY,
            lambda: self.__update_width_dependent_margins(),
        )

        controller = Gtk.EventControllerKey()
        controller.connect("key-pressed", self.__on_key_pressed)
        controller.connect("key-released", self.__on_key_released)
        self.add_controller(controller)

        controller = Gtk.EventControllerMotion()
        controller.connect("motion", self.__on_motion)
        controller.connect("leave", self.__on_motion_leave)
        self.add_controller(controller)

        self.__spellchecker = None
        if self.__config_manager.spelling_enabled:
            self.__init_spellchecker()

        self.__config_manager.connect_changed(
            ConfigManager.SPELLING_ENABLED,
            self.__on_spelling_toggled,
        )

        # TODO Temporary hack to provide access to spelling on mobile
        self.__setup_longpress_touch_menu()

    def setup(self) -> None:
        """Initialisation after window creation."""
        window = self.get_root()
        window.connect("notify::fullscreened", lambda _o, _v: self.__update_all_margins())

    def do_size_allocate(self, width: int, height: int, baseline: int) -> None:
        """Allocates widget with a transformation that translates the origin to the position in
        allocation.

        :param int width: Width of the allocation
        :param int height: Height of the allocation
        :param int baseline: The baseline of the child
        """
        GtkSource.View.do_size_allocate(self, width, height, baseline)
        if width != self.__width_margins_set_for:
            self.__update_width_dependent_margins()
        if height != self.__height_margins_set_for:
            self.__update_bottom_margin(height)

    def jump_to_insertion_point(self) -> None:
        """Jump to the insertion point.

        This jump is used to avoid issues with height allocation in `scroll_to_mark` when the
        buffer contains lines of varying height.
        """
        buffer = self.get_buffer()
        insert_iter = buffer.get_iter_at_mark(buffer.get_insert())
        location = self.get_iter_location(insert_iter)
        adjustment = self.get_vadjustment()

        # Push the insertion point down the view a little so that when the replace headerbar is
        # visible the point should still be below that.
        adjustment.set_value(location.y - self.get_height() / 4)

    def update_font_family(self, font_family: str) -> None:
        """Update font family from system preference.

        :param str font_family: The font family
        """
        self.__font_family = font_family
        self.__update_font()

    @GObject.Property(type=bool, default=False)
    def spellchecker_enabled(self) -> bool:
        if self.__spellchecker is not None:
            return self.__spellchecker.enabled
        else:
            return False

    @spellchecker_enabled.setter
    def set_spellchecker_enabled(self, value: bool) -> None:
        if self.__spellchecker is not None:
            self.__spellchecker.enabled = value

    @GObject.Property(type=int, default=-1)
    def line_length(self) -> int:
        return self.__line_length

    @line_length.setter
    def set_line_length(self, value: int) -> None:
        self.__line_length = value
        # Ensure margin resizing works during resize (but only if we're initialised)
        if self.__width_margins_set_for:
            self.__update_width_dependent_margins()

    @GObject.Property(type=int, default=0)
    def header_bar_height(self) -> int:
        return self.__header_bar_height

    @header_bar_height.setter
    def set_header_bar_height(self, value: int) -> None:
        self.__header_bar_height = value

    @GObject.Property(type=int, default=0)
    def padded_header_bar_height(self) -> int:
        return self.__header_bar_height + self.MINIMUM_MARGIN

    @GObject.Property(type=bool, default=False)
    def accommodate_double_header_bar(self) -> bool:
        return self.__accommodate_double_header_bar

    @accommodate_double_header_bar.setter
    def set_accommodate_double_header_bar(self, value: bool) -> None:
        self.__accommodate_double_header_bar = value
        if self.__width_margins_set_for:
            self.__update_width_dependent_margins()

    # TODO Part of temporary hack to provide access to spelling menu on mobile
    def __on_longpress(self, _gesture: Gtk.GestureLongPress, x: float, y: float) -> None:
        # Long press menu is only currently used for spelling on mobile, don't show many if
        # spelling disabled
        if self.__spellchecker is None or not self.__spellchecker.enabled:
            return
        buffer_x, buffer_y = self.window_to_buffer_coords(Gtk.TextWindowType.TEXT, int(x), int(y))
        iter = self.get_iter_at_location(buffer_x, buffer_y)[1]
        self.__spellchecker.move_click_mark(iter)
        self.__spellchecker.populate_menu(self.__longpress_popover.get_menu_model())
        rect = Gdk.Rectangle()
        rect.x = x
        rect.y = y
        rect.width = rect.height = 1
        self.__longpress_popover.set_pointing_to(rect)
        self.__longpress_popover.popup()

    def __on_spelling_toggled(self) -> None:
        if self.__config_manager.spelling_enabled:
            self.__init_spellchecker()
            self.spellchecker_enabled = True
        else:
            self.spellchecker_enabled = False

    def __on_key_pressed(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        buffer = self.get_buffer()
        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter, Gdk.KEY_ISO_Enter):
            if list_formatter.check_and_extend_list(buffer):
                # Attempt to track cursor in viewport
                mark = buffer.get_insert()
                insert_iter = buffer.get_iter_at_mark(mark)
                buffer_coords = self.get_iter_location(insert_iter)
                visible_rect = self.get_visible_rect()
                if buffer_coords.y + buffer_coords.height >= visible_rect.y + visible_rect.height:
                    # Without adding as an idle task we sometimes end up with an insufficient scroll
                    GLib.idle_add(self.emit, "move-viewport", Gtk.ScrollStep.STEPS, 1)

                return Gdk.EVENT_STOP
        elif keyval in (Gdk.KEY_Tab, Gdk.KEY_ISO_Left_Tab):
            if state == Gdk.ModifierType.SHIFT_MASK:
                list_formatter.decrease_indentation(buffer)
            else:
                list_formatter.increase_indentation(buffer)
            return Gdk.EVENT_STOP
        elif keyval in (Gdk.KEY_Control_L, Gdk.KEY_Control_R):
            if not self.__control_down:
                self.__control_down = True
                self.__refresh_mouse_cursor()

        return Gdk.EVENT_PROPAGATE

    def __on_key_released(
        self,
        controller: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval in (Gdk.KEY_Control_L, Gdk.KEY_Control_R):
            if self.__control_down:
                self.__control_down = False
                self.__refresh_mouse_cursor()

        return Gdk.EVENT_PROPAGATE

    def __on_motion(
        self,
        controller: Gtk.EventControllerMotion,
        x: float,
        y: float,
    ) -> bool:
        self.__mouse_x, self.__mouse_y = self.window_to_buffer_coords(
            Gtk.TextWindowType.WIDGET, int(x), int(y)
        )
        self.__refresh_mouse_cursor()
        return Gdk.EVENT_PROPAGATE

    def __on_motion_leave(
        self,
        controller: Gtk.EventControllerMotion,
    ) -> bool:
        self.__mouse_x = None
        self.__mouse_y = None
        self.__refresh_mouse_cursor()
        return Gdk.EVENT_PROPAGATE

    def __refresh_mouse_cursor(self) -> None:
        show_cursor = False
        if self.__control_down and self.__mouse_x is not None:
            success, iter = self.get_iter_at_location(self.__mouse_x, self.__mouse_y)
            if success:
                buffer = self.get_buffer()
                classes = set(buffer.get_context_classes_at_iter(iter))
                show_cursor = len(classes.intersection(self.LINK_TAGS)) > 0

        if self.__showing_pointer_cursor != show_cursor:
            self.set_cursor(self.__cursor_pointer if show_cursor else None)
            self.__showing_pointer_cursor = show_cursor

    def __update_all_margins(self) -> None:
        self.__update_width_dependent_margins()
        self.__update_bottom_margin(self.get_height())

    def __update_width_dependent_margins(self) -> None:
        width_range_progress = self.__get_width_margin_range_progress()
        self.__update_top_margin(width_range_progress)
        self.__update_side_margins(width_range_progress)
        self.emit("margins-updated")

    def __get_width_margin_range_progress(self) -> float:
        width = self.get_width()
        minimum_width_for_margin = 400
        max_width_for_margin = self.__line_length if self.__line_length > 500 else 800

        if width <= minimum_width_for_margin:
            progress = 0.0
        elif width < max_width_for_margin:
            value_range = max_width_for_margin - minimum_width_for_margin
            in_range = self.get_width() - minimum_width_for_margin
            progress = in_range / value_range
        else:
            progress = 1.0
        return progress

    def __update_top_margin(self, range_progress: float) -> None:
        width = self.get_width()
        self.__width_margins_set_for = width

        y_margin_base = self.__header_bar_height

        y_range = float(self.__header_bar_height - self.MINIMUM_MARGIN)
        y_margin_base = self.MINIMUM_MARGIN + int(range_progress * y_range)

        if self.__accommodate_double_header_bar:
            y_margin_base = max(y_margin_base, self.__header_bar_height)

        # If the header bar is non hiding (set visible for the fullscreen state) there will be
        # padding added in the Editor's __update_toolbar_underlay_margins to ensure the text avoids
        # the headerbar. Otherwise we add it here.
        if self.__config_manager.editor_header_bar_visible_for_window_state:
            y_margin = y_margin_base
        else:
            y_margin = y_margin_base + self.padded_header_bar_height

        if self.props.top_margin != y_margin:
            self.set_top_margin(y_margin)

    def __update_bottom_margin(self, height: int) -> None:
        self.__height_margins_set_for = height
        new_margin = int(height * 0.4)
        if self.props.bottom_margin != new_margin:
            self.set_bottom_margin(new_margin)

    def __update_side_margins(self, range_progress: float) -> None:
        width = self.get_width()
        if width < 1:
            return

        if self.__line_length > 0 and width - 2 * self.STANDARD_X_MARGIN > self.__line_length:
            x_margin = (width - self.__line_length) / 2
        else:
            x_range = float(self.STANDARD_X_MARGIN - self.MINIMUM_MARGIN)
            x_margin = self.MINIMUM_MARGIN + int(range_progress * x_range)

        if self.get_left_margin() != x_margin:
            self.set_left_margin(x_margin)
            self.set_right_margin(x_margin)

    def __update_font(self) -> None:
        if not self.__css_provider:
            self.__css_provider = Gtk.CssProvider()
            self.get_style_context().add_provider(
                self.__css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER
            )

        size = self.__config_manager.font_size
        style = f"""
        .editor-textview {{
            font-size: {size}pt;
            font-family: {self.__font_family}, monospace;
        }}"""
        self.__css_provider.load_from_data(style, -1)

    def __init_spellchecker(self) -> None:
        if self.__spellchecker is not None:
            return

        pref_language = self.__config_manager.spelling_language
        if pref_language is not None:
            language = pref_language
            logging.debug(f'Attempting to use spelling language from preference "{pref_language}"')
        else:
            default_language, _encoding = locale.getdefaultlocale()
            if default_language:
                language = default_language
                logging.debug(f'Attempting to use locale default spelling language "{language}"')
            else:
                logging.warning("Couldn't determine locale default language")

        try:
            self.__spellchecker = SpellChecker(self, language, collapse=False)
        except NoDictionariesFound:
            # Not communicating via UI for now as systems without any dictionaries are likely to
            # be corner cases, bespoke systems eg. Arch etc
            logging.warning("Disabling spell checker as no dictionaries were found")
            return

        assert self.__spellchecker  # mypy

        self.spellchecker_enabled = False
        self.__spellchecker.batched_rechecking = True
        if pref_language is not None:
            self.__verify_preferred_language_in_use(pref_language)
        self.__spellchecker.connect(
            "notify::language", lambda _o, _v: self.__spelling_language_changed()
        )

        buffer = self.get_buffer()
        table = buffer.get_tag_table()

        def process_added_tag(tag):
            if tag.get_property("name") == self.SOURCEVIEW_NO_SPELLCHECK_TAG:
                self.__spellchecker.append_ignore_tag(self.SOURCEVIEW_NO_SPELLCHECK_TAG)

        def tag_added(tag, *args):
            if isinstance(tag, Gtk.TextTag):
                process_added_tag(tag)
            elif len(args) > 0 and isinstance(args[0], Gtk.TextTag):
                process_added_tag(args[0])

        table.connect("tag-added", tag_added)

    def __verify_preferred_language_in_use(self, pref_language: str) -> None:
        assert self.__spellchecker
        language_in_use = self.__spellchecker.language
        if language_in_use != pref_language:
            logging.warning(
                f'Spelling language from preference "{pref_language}" not found, clearing'
                " preference"
            )

            logging.info("Available languages:")
            for code, name in self.__spellchecker.languages:
                logging.info(" - %s (%5s)" % (name, code))

            self.__config_manager.spelling_language = ""

    def __spelling_language_changed(self) -> None:
        assert self.__spellchecker
        language = self.__spellchecker.language
        logging.info(f'New spelling language "{language}"')
        self.__config_manager.spelling_language = language

    # TODO Part of temporary hack to provide access to spelling menu on mobile
    def __setup_longpress_touch_menu(self) -> None:
        gesture = Gtk.GestureLongPress.new()
        gesture.set_touch_only(True)

        gesture.connect("pressed", self.__on_longpress)

        self.add_controller(gesture)

        self.__longpress_popover = Gtk.PopoverMenu.new_from_model(Gio.Menu())
        self.__longpress_popover.set_parent(self)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GLib, GObject, Gtk, GtkSource, Pango

from typing import Callable, Optional

from iotas.formatter import Formatter, LinkStateInfo
from iotas.link_dialog import LinkDialog
from iotas.table_dialog import TableDialog


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/formatting_header_bar.ui")
class FormattingHeaderBar(Adw.BreakpointBin):
    """Editor formatting header bar."""

    __gtype_name__ = "FormattingHeaderBar"
    __gsignals__ = {
        "height-change": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    _breakpoint: Adw.Breakpoint = Gtk.Template.Child()
    _base_box: Gtk.Box = Gtk.Template.Child()
    _parent_box: Gtk.Box = Gtk.Template.Child()
    _extend_box: Gtk.Box = Gtk.Template.Child()
    _heading_menu: Gtk.MenuButton = Gtk.Template.Child()
    _menu_button: Gtk.MenuButton = Gtk.Template.Child()
    _header_bar: Adw.HeaderBar = Gtk.Template.Child()
    _link_button: Gtk.Button = Gtk.Template.Child()
    _table_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self._breakpoint.connect("apply", lambda _o: self.__apply_breakpoint())
        self._breakpoint.connect("unapply", lambda _o: self.__unapply_breakpoint())
        self.__formatter = Formatter()
        self.__link_dialog: LinkDialog | None = None
        self.__table_dialog: TableDialog | None = None
        self.__active = False

    def setup(self, sourceview: GtkSource.View) -> None:
        """Later startup initialisation.

        :param GtkSource.View sourceview: Sourceview
        """
        self.__setup_actions()
        self.__sourceview = sourceview

    def prepare(self) -> None:
        """Initialise state entering editor."""
        self.__formatter.buffer = self.__sourceview.get_buffer()

        def callback():
            if self.get_width() == 0:
                GLib.idle_add(callback)
                return

            if self.get_width() < 500:
                self.__make_shorter()
            else:
                self.__make_taller()

        GLib.idle_add(callback)

    def focus(self) -> None:
        """Grab focus."""
        self._heading_menu.grab_focus()

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @active.setter
    def set_active(self, value: bool) -> None:
        self.__active = value
        self.__enable_actions(value)

    def __on_table_create(self, _obj: TableDialog, width: int, height: int) -> None:
        view_width = self.__sourceview.get_width()
        if self.__sourceview.line_length > 0:
            view_width = min(view_width, self.__sourceview.line_length)
        if self.__get_text_direction() == Gtk.TextDirection.RTL:
            pango_context = self.__sourceview.get_rtl_context()
        else:
            pango_context = self.__sourceview.get_ltr_context()
        metrics = pango_context.get_metrics()
        character_width = metrics.get_approximate_char_width()
        self.__formatter.table(width, height, view_width, Pango.units_to_double(character_width))

        # Can't directly call grab_focus from idle_add otherwise its returning of True will retain,
        # and repeat, the event
        def focus_sourceview():
            self.__sourceview.grab_focus()

        GLib.idle_add(focus_sourceview)

    def __on_link_apply(self, _obj: LinkDialog, info: LinkStateInfo) -> None:
        self.__formatter.link(info)
        self.__sourceview.grab_focus()

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        def wrap(func: Callable, param: Optional[int] = None) -> None:
            args = (param,) if param is not None else ()
            func(*args)
            self.__sourceview.grab_focus()

        action = Gio.SimpleAction.new("bold")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.bold))
        action_group.add_action(action)
        app.set_accels_for_action("format.bold", ["<Control>b"])

        action = Gio.SimpleAction.new("italic")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.italic))
        action_group.add_action(action)
        app.set_accels_for_action("format.italic", ["<Control>i"])

        action = Gio.SimpleAction.new("strikethrough")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.strikethrough))
        action_group.add_action(action)
        app.set_accels_for_action("format.strikethrough", ["<Control><Shift>x"])

        action = Gio.SimpleAction.new("quote")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.quote))
        action_group.add_action(action)
        app.set_accels_for_action("format.quote", ["<Control><Shift>q"])

        action = Gio.SimpleAction.new("heading-1")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 1))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-1", ["<Control>1"])

        action = Gio.SimpleAction.new("heading-2")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 2))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-2", ["<Control>2"])

        action = Gio.SimpleAction.new("heading-3")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 3))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-3", ["<Control>3"])

        action = Gio.SimpleAction.new("heading-4")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 4))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-4", ["<Control>4"])

        action = Gio.SimpleAction.new("heading-5")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 5))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-5", ["<Control>5"])

        action = Gio.SimpleAction.new("heading-6")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 6))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-6", ["<Control>6"])

        action = Gio.SimpleAction.new("heading-remove")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.heading, 0))
        action_group.add_action(action)
        app.set_accels_for_action("format.heading-remove", ["<Control>7"])

        action = Gio.SimpleAction.new("unordered-list")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.unordered_list))
        action_group.add_action(action)
        app.set_accels_for_action("format.unordered-list", ["<Control>u"])

        action = Gio.SimpleAction.new("ordered-list")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.ordered_list))
        action_group.add_action(action)
        app.set_accels_for_action("format.ordered-list", ["<Control>o"])

        action = Gio.SimpleAction.new("checkbox")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.checkbox))
        action_group.add_action(action)
        app.set_accels_for_action("format.checkbox", ["<Alt>c"])

        action = Gio.SimpleAction.new("toggle-checkbox")
        action.connect("activate", lambda _a, _p: self.__formatter.toggle_checkbox())
        action_group.add_action(action)
        app.set_accels_for_action("format.toggle-checkbox", ["<Control>t"])

        action = Gio.SimpleAction.new("link")
        action.connect("activate", lambda _a, _p: self.__link())
        action_group.add_action(action)
        app.set_accels_for_action("format.link", ["<Control>k"])

        action = Gio.SimpleAction.new("code")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.code))
        action_group.add_action(action)
        app.set_accels_for_action("format.code", ["<Control><Shift>c"])

        action = Gio.SimpleAction.new("horizontal-rule")
        action.connect("activate", lambda _a, _p: wrap(self.__formatter.horizontal_rule))
        action_group.add_action(action)
        app.set_accels_for_action("format.horizontal-rule", ["<Control><Shift>r"])

        action = Gio.SimpleAction.new("table")
        action.connect("activate", lambda _a, _p: self.__table())
        action_group.add_action(action)
        app.set_accels_for_action("format.table", ["<Control><Shift>t"])

        app.get_active_window().insert_action_group("format", action_group)
        self.__action_group = action_group

    def __enable_actions(self, enabled: bool) -> None:
        actions = self.__action_group.list_actions()
        for action in actions:
            self.__action_group.lookup_action(action).set_enabled(enabled)

    def __apply_breakpoint(self) -> None:
        self.__make_shorter()
        self.emit("height-change")

    def __make_shorter(self) -> None:
        self._extend_box.set_visible(False)
        self._menu_button.set_visible(True)
        self._base_box.set_spacing(4)
        self._parent_box.set_spacing(4)
        self._header_bar.add_css_class("formatting-small")
        self.set_property("height-request", 24)

    def __unapply_breakpoint(self) -> None:
        self.__make_taller()
        self.emit("height-change")

    def __make_taller(self) -> None:
        self._extend_box.set_visible(True)
        self._menu_button.set_visible(False)
        self._base_box.set_spacing(6)
        self._parent_box.set_spacing(6)
        self._header_bar.remove_css_class("formatting-small")
        self.set_property("height-request", 47)

    def __link(self) -> None:
        """Show link dialog."""
        info = self.__formatter.get_link_state()
        if not self.__link_dialog:
            self.__link_dialog = LinkDialog()
            self.__link_dialog.connect("apply", self.__on_link_apply)
        self.__link_dialog.show(info, self.get_root())

    def __table(self) -> None:
        """Show table dialog."""
        if not self.__table_dialog:
            self.__table_dialog = TableDialog()
            self.__table_dialog.connect("create", self.__on_table_create)
        self.__table_dialog.show(self.get_root())

    def __get_text_direction(self) -> Gtk.TextDirection:
        widget_direction = self.__sourceview.get_direction()
        if widget_direction:
            direction = widget_direction
        else:
            direction = Gtk.Widget.get_default_direction()
        return direction



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, Gtk

from iotas.config_manager import ConfigManager


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/theme_selector.ui")
class ThemeSelector(Gtk.Box):
    __gtype_name__ = "ThemeSelector"

    _follow: Gtk.CheckButton = Gtk.Template.Child()
    _light: Gtk.CheckButton = Gtk.Template.Child()
    _dark: Gtk.CheckButton = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()
        self.__config_manager = ConfigManager.get_default()
        self.__populate()
        self.__config_manager.connect_changed(ConfigManager.STYLE, self.__populate)

    @Gtk.Template.Callback()
    def _on_option_selected(self, _widget: Gtk.CheckButton) -> None:
        name = None
        if self._follow.get_active():
            name = "follow"
        elif self._light.get_active():
            name = "light"
        elif self._dark.get_active():
            name = "dark"
        # Name can be null when checkbutton is unchecked for radio group. A positive check will
        # follow.
        if name and self.__config_manager.style != name:
            self.__config_manager.style = name
            app = Gio.Application.get_default()
            app.apply_style()

    def __populate(self) -> None:
        style = self.__config_manager.style
        if style == "follow":
            self._follow.set_active(True)
        elif style == "light":
            self._light.set_active(True)
        else:
            self._dark.set_active(True)



# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, GObject, Gtk, GtkSource

from typing import Optional

from iotas.editor_search_entry import EditorSearchEntry


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/editor_search_header_bar.ui")
class EditorSearchHeaderBar(Gtk.Box):
    __gtype_name__ = "EditorSearchHeaderBar"
    __gsignals__ = {
        "resumed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "open-for-replace": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    _search_entry: EditorSearchEntry = Gtk.Template.Child()
    _backward_button: Gtk.Button = Gtk.Template.Child()
    _forward_button: Gtk.Button = Gtk.Template.Child()
    _revealer: Gtk.Revealer = Gtk.Template.Child()
    _replace_entry: Gtk.Entry = Gtk.Template.Child()
    _replace_toggle: Gtk.ToggleButton = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self.__settings = GtkSource.SearchSettings.new()
        self.__settings.set_wrap_around(True)
        self._search_entry.text.bind_property(
            "text",
            self.__settings,
            "search-text",
        )
        self.__context: Optional[GtkSource.SearchContext] = None
        self.__offset_when_entered = 0
        self.__context_signal_handler_id: Optional[int] = None
        self.__cursor_signal_handler_id: Optional[int] = None
        self.__current_match_tag = None
        self.__avoid_jumping_during_replace = False
        self.__active = False
        self.__replace_visible = False
        self._search_entry.text.connect(
            "notify::text", lambda _o, _v: self.__on_search_text_changed()
        )

        self._revealer.bind_property(
            "child-revealed", self._replace_toggle, "active", GObject.BindingFlags.SYNC_CREATE
        )

    def setup(self, sourceview: GtkSource.View):
        """Perform initial setup."""
        self.__setup_actions()
        self.__sourceview = sourceview

    def enter(self, resuming: bool, for_replace: bool) -> None:
        """Enter search.

        :param bool resuming: Whether resuming search
        :param bool for_replace: Whether opening directly from a ctrl-h shortcut
        """
        buffer = self.__sourceview.get_buffer()
        self.__context = GtkSource.SearchContext.new(buffer, self.__settings)
        insert_iter = buffer.get_iter_at_mark(buffer.get_insert())
        self.__offset_when_entered = insert_iter.get_offset()
        self.__setup_current_match_styling()
        if for_replace:
            self.replace_visible = True
            self._revealer.set_reveal_child(True)

        # Ensure when restarting search any term remains selected so it can be easily replaced by
        # typing.
        self.__restarting_with_nonempty_term = not resuming and self._search_entry.get_text() != ""
        if self.__restarting_with_nonempty_term:
            self._search_entry.show_working()

        self._search_entry.text.grab_focus()
        if resuming:
            self._search_entry.text.set_position(-1)
        elif self._search_entry.get_text() != "":
            self._search_entry.text.select_region(0, -1)

        self.__context_signal_handler_id = self.__context.connect(
            "notify::occurrences-count", lambda _o, _v: self.__on_occurrences_count_changed()
        )
        self.__cursor_signal_handler_id = buffer.connect("cursor-moved", self.__on_cursor_moved)
        self.__replace_action.set_enabled(False)
        self.__active = True

    def exit(self) -> None:
        """Exit search."""
        assert self.__context
        buffer = self.__sourceview.get_buffer()
        buffer.disconnect(self.__cursor_signal_handler_id)
        self.__context.disconnect(self.__context_signal_handler_id)
        self.__context = None
        begin, end = buffer.get_bounds()
        buffer.remove_tag(self.__current_match_tag, begin, end)
        self.__current_match_tag = None
        self._revealer.set_reveal_child(False)
        self.replace_visible = False
        self.__active = False

    def disable_actions(self) -> None:
        """Disable actions."""
        self.__action_group.lookup_action("backward").set_enabled(False)
        self.__action_group.lookup_action("forward").set_enabled(False)

    def refocus_search_and_select(self) -> None:
        """Grab focus to search entry and select any text."""
        self._search_entry.text.grab_focus()
        if self._search_entry.get_text() != "":
            self._search_entry.text.select_region(0, -1)

    @GObject.Property(type=bool, default=False)
    def active(self) -> bool:
        return self.__active

    @GObject.Property(type=bool, default=False)
    def replace_visible(self) -> bool:
        return self.__replace_visible

    @replace_visible.setter
    def set_replace_visible(self, value: bool) -> None:
        self.__replace_visible = value

    @Gtk.Template.Callback()
    def _on_search_entry_activate(self, _obj: GObject.Object) -> None:
        self.__move_forward()

    @Gtk.Template.Callback()
    def _on_replace_entry_activate(self, _obj: GObject.Object) -> None:
        if self.__replace_action.get_enabled():
            self.__replace_selection()

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        app = Gio.Application.get_default()

        action = Gio.SimpleAction.new("forward")
        action.connect("activate", lambda _a, _v: self.__move_forward())
        action_group.add_action(action)
        app.set_accels_for_action("editor-search.forward", ["<Control>g"])
        action.set_enabled(False)

        action = Gio.SimpleAction.new("backward")
        action.connect("activate", lambda _a, _v: self.__move_backward())
        action_group.add_action(action)
        app.set_accels_for_action("editor-search.backward", ["<Shift><Control>g"])
        action.set_enabled(False)

        action = Gio.SimpleAction.new("toggle-replace")
        action.connect("activate", lambda _a, _p: self.__toggle_replace_visible())
        action_group.add_action(action)
        app.set_accels_for_action("editor-search.toggle-replace", ["<Control>h"])

        action = Gio.SimpleAction.new("replace")
        action.connect("activate", lambda _a, _p: self.__replace_selection())
        action_group.add_action(action)
        self.__replace_action = action

        self.__action_group = action_group
        app.get_active_window().insert_action_group("editor-search", action_group)

    def __on_context_forward(self, _obj: GObject.Object, result: Gio.AsyncResult) -> None:
        assert self.__context
        assert self.__current_match_tag
        success, match_start, match_end, __ = self.__context.forward_finish(result)
        if success:
            buffer = self.__sourceview.get_buffer()
            if self.__restarting_with_nonempty_term:
                buffer.place_cursor(match_start)
            else:
                buffer.select_range(match_start, match_end)
                self.__current_match_tag.set_priority(buffer.get_tag_table().get_size() - 1)
                self.__update_for_current_match(match_start, match_end)
            self.__sourceview.jump_to_insertion_point()
        if self.__restarting_with_nonempty_term:
            self.__restarting_with_nonempty_term = False
        self.__replace_action.set_enabled(success)

    def __on_context_backward(self, _obj: GObject.Object, result: Gio.AsyncResult) -> None:
        assert self.__current_match_tag
        success, match_start, match_end, __ = self.__context.backward_finish(result)
        if success:
            buffer = self.__sourceview.get_buffer()
            buffer.select_range(match_start, match_end)
            self.__current_match_tag.set_priority(buffer.get_tag_table().get_size() - 1)
            self.__sourceview.jump_to_insertion_point()
            self.__update_for_current_match(match_start, match_end)
        self.__replace_action.set_enabled(success)

    def __on_cursor_moved(self, _obj: GObject.Object) -> None:
        self.__update_for_current_match(None, None)
        self.__replace_action.set_enabled(False)

    def __on_search_text_changed(self) -> None:
        self.__replace_action.set_enabled(False)

        # Clear any tagged matches when deleting the search term
        buffer = self.__sourceview.get_buffer()
        begin, end = buffer.get_bounds()
        buffer.remove_tag(self.__current_match_tag, begin, end)

    def __on_occurrences_count_changed(self) -> None:
        assert self.__context
        count = self.__context.get_occurrences_count()
        self._search_entry.set_occurrence_count(count)
        self._search_entry.set_occurrence_position(0)
        search_can_move = count > 0

        if not self.__sourceview.has_focus():
            if search_can_move and not self.__avoid_jumping_during_replace:
                self.__jump_to_first()
            elif self.__restarting_with_nonempty_term:
                self.__restarting_with_nonempty_term = False

        self.__action_group.lookup_action("backward").set_enabled(search_can_move)
        self.__action_group.lookup_action("forward").set_enabled(search_can_move)
        self.__avoid_jumping_during_replace = False

    def __jump_to_first(self) -> None:
        buffer = self.__sourceview.get_buffer()
        start_iter = buffer.get_iter_at_offset(self.__offset_when_entered)
        self.__move_forward(start_iter)

    def __move_forward(self, from_iter: Optional[Gtk.TextIter] = None) -> None:
        """Move to next search match."""
        if self.__context is None:
            if self._search_entry.get_text() != "":
                self.emit("resumed")
            return
        if not from_iter:
            buffer = self.__sourceview.get_buffer()
            if buffer.get_has_selection():
                begin, end = buffer.get_selection_bounds()
                begin.order(end)
                from_iter = end
            else:
                mark = buffer.get_insert()
                from_iter = buffer.get_iter_at_mark(mark)
        self.__context.forward_async(from_iter, None, self.__on_context_forward)

    def __move_backward(self) -> None:
        """Move to previous search match."""
        if self.__context is None:
            if self._search_entry.get_text() != "":
                self.emit("resumed")
            return
        buffer = self.__sourceview.get_buffer()
        if buffer.get_has_selection():
            begin, end = buffer.get_selection_bounds()
            begin.order(end)
        else:
            mark = buffer.get_insert()
            begin = buffer.get_iter_at_mark(mark)
        self.__context.backward_async(begin, None, self.__on_context_backward)

    def __toggle_replace_visible(self) -> None:
        if not self.__active:
            self.emit("open-for-replace")
            return
        reveal = not self._revealer.get_child_revealed()
        self._revealer.set_reveal_child(reveal)
        self.replace_visible = reveal
        if reveal:
            # Clearing the replace entry here to avoid the current match selection in the buffer
            # being removed, which would result in replace not working until the match is
            # re-navigated to. Ideally selection wouldn't be used to identify the current search
            # match.
            self._replace_entry.set_text("")
            self._replace_entry.grab_focus()
            self.__replace_action.set_enabled(self.__have_search_term_selection())

    def __update_for_current_match(
        self, match_start: Optional[Gtk.TextIter], match_end: Optional[Gtk.TextIter]
    ) -> None:
        assert self.__context
        assert self.__current_match_tag
        buffer = self.__sourceview.get_buffer()

        new_position = 0
        if buffer.get_has_selection():
            bounds = buffer.get_selection_bounds()
            if len(bounds) == 2:
                begin, end = bounds
                begin.order(end)
                new_position = self.__context.get_occurrence_position(begin, end)
        self._search_entry.set_occurrence_position(new_position)

        if self.__current_match_tag is not None:
            begin, end = buffer.get_bounds()
            buffer.remove_tag(self.__current_match_tag, begin, end)
        if match_start is not None and match_end is not None:
            buffer.apply_tag(self.__current_match_tag, match_start, match_end)
            self.__current_match_tag.set_priority(buffer.get_tag_table().get_size() - 1)

    def __setup_current_match_styling(self) -> None:
        buffer = self.__sourceview.get_buffer()
        table = buffer.get_tag_table()
        scheme = buffer.get_style_scheme()

        if self.__current_match_tag is not None:
            table.remove(self.__current_match_tag)
        self.__current_match_tag = buffer.create_tag()
        assert self.__current_match_tag

        if scheme is not None:
            style = scheme.get_style("current-search-match")
            if style is not None:
                style.apply(self.__current_match_tag)
        self.__current_match_tag.set_priority(table.get_size() - 1)

    def __replace_selection(self) -> None:
        buffer = self.__sourceview.get_buffer()
        if not self.__have_search_term_selection():
            return
        start, end = buffer.get_selection_bounds()
        self.__avoid_jumping_during_replace = True
        buffer.delete(start, end)
        new_text = self._replace_entry.get_text()
        if new_text != "":
            buffer.insert(start, new_text)
        self.__move_forward()

    def __have_search_term_selection(self) -> bool:
        buffer = self.__sourceview.get_buffer()
        if not buffer.get_has_selection():
            return False
        start, end = buffer.get_selection_bounds()
        return start.get_text(end).lower() == self._search_entry.get_text().lower()
